import abc
import math
import collections
from functools import cache, cached_property

import numpy as np
from scipy.signal import welch
from astropy import units as u

from arte.time_series.axis_handler import AxisHandler
from arte.utils.not_available import NotAvailable
from arte.utils.help import add_help, modify_help
from arte.utils.unit_checker import make_sure_its_a


class TimeSeriesException(Exception):
    '''Exception raised by TimeSeries'''


@add_help
class TimeSeries(metaclass=abc.ABCMeta):
    '''
    Base class for implementing operations on data representing time series.

    Derived classes must implement a `_get_not_indexed_data()` method
    that returns a numpy array of shape (n_time_elements, n_ensemble_elements).

    Derived classes must also implement a `get_index_of()` method to add
    ensemble indexing with arbitrary *args and **kwargs parameters
    (e.g. returning of partial subset based on indexes or names).

    Originally implemented as part of the ARGOS codebase.
    '''

    def __init__(self, axes=None):
        self._frequency = None
        self._last_cut_frequency = None
        self._power = None
        self._segment_factor = None
        self._window = None
        self._axis_handler = AxisHandler(axes)

    @abc.abstractmethod
    def _get_not_indexed_data(self):
        pass

    @property
    def axes(self):
        '''Data series shape'''
        return self._axis_handler.axes()

    @cache
    def _get_time_vector(self):
        '''Override to provide a custom time vector'''
        return np.arange(len(self._get_not_indexed_data()))
    
    def _get_time_vector_check(self):
        v = self._get_time_vector()
        if isinstance(v, np.ndarray):
            return v
        else:
            try:
                # Try to convert
                return np.array(v)
            except TypeError as e:
                raise TimeSeriesException('Cannot convert _get_time_vector() result to numpy array') from e

    def get_data(self, *args, times=None, axes=None, **kwargs):
        '''Raw data as a matrix [time, series]'''

        data = self._get_not_indexed_data()
        if times is not None:
            if not isinstance(times, collections.abc.Sequence) or len(times) != 2:
                raise TimeSeriesException('Times keywords must be a sequence of two elements: [start, stop]')
            time_vector = self.get_time_vector()
            if len(time_vector) != len(data):
                raise TimeSeriesException('Time vector and data lengths differ')

            start, stop = times
            if isinstance(time_vector, u.Quantity):
                if start is not None:
                    start = make_sure_its_a(time_vector.unit, start)
                if stop is not None:
                    stop = make_sure_its_a(time_vector.unit, stop)

            idxs = np.ones(len(time_vector), dtype=bool)
            if start is not None:
                idxs = np.logical_and(idxs, time_vector >= start)
            if stop is not None:
                idxs = np.logical_and(idxs, time_vector < stop)
            data = data[idxs]

        index = self.get_index_of(*args, **kwargs)
        data = self._index_data(data, index)
        data = self._axis_handler.transpose(data, axes)
        return data

    def _index_data(self, data, index):
        if index is None:
            return data
        elif isinstance(index, tuple):
            return data[(slice(0,data.shape[0]),) + index]
        else:
            return data[:, index]

    @abc.abstractmethod
    def get_index_of(self, *args, **kwargs):
        pass

    def data_unit(self):
        '''Override to return a string with a compact unit notation'''
        return None

    def data_label(self):
        '''Override to return a string with a readable unit name'''
        return None

    @cached_property
    def delta_time(self):
        '''Property with the interval between samples.
        
        If no interval can be determined (time vector too short),
        returns 1 with the correct unit if applicable.
        '''
        time_vector = self.get_time_vector()
        if len(time_vector) == 0:
            return 1
        diff = np.diff(time_vector)
        if len(diff) > 0:
            return np.median(diff)
        else:
            if isinstance(time_vector[0], u.Quantity):
                return 1 * time_vector[0].unit
            else:
                return 1

    def frequency(self):
        return self._frequency

    def get_time_vector(self):
        '''Return the series time vector'''
        return self._get_time_vector_check()

    def last_cut_frequency(self):
        return self._last_cut_frequency

    @cache
    def ensemble_size(self):
        '''Number of distinct series in this time ensemble'''
        not_indexed_data = self._get_not_indexed_data()
        return math.prod(not_indexed_data.shape[1:])

    @cache
    def time_size(self):
        '''Number of time samples in this time ensemble'''
        not_indexed_data = self._get_not_indexed_data()
        return not_indexed_data.shape[0]

    def _data_flattened(self, data):
        '''Flatten data over all non-time dimensions'''
        return data.reshape(data.shape[0], self._data_sample_size(data))

    def _data_sample_shape(self, data):
        return data.shape[1:]

    def _data_sample_axes(self, data):
        return tuple(range(len(data.shape))[1:])

    def _data_sample_size(self, data):
        return math.prod(data.shape[1:])

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def time_median(self, *args, times=None, **kwargs):
        '''Median over time for each series'''
        return np.median(self.get_data(*args, times=times, **kwargs), axis=0)

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def time_std(self, *args, times=None, **kwargs):
        '''Standard deviation over time for each series'''
        return np.std(self.get_data(*args, times=times, **kwargs), axis=0)

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def time_average(self, *args, times=None, **kwargs):
        '''Average value over time for each series'''
        return np.mean(self.get_data(*args, times=times, **kwargs), axis=0)

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def time_rms(self, *args, times=None, **kwargs):
        '''Root-Mean-Square value over time for each series'''
        x = self.get_data(*args, times=times, **kwargs)
        return np.sqrt(np.mean(np.abs(x)**2, axis=0))


    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def ensemble_average(self, *args, times=None, **kwargs):
        '''Average across series at each sampling time'''
        data = self.get_data(*args, times=times, **kwargs)
        return np.mean(data, axis=self._data_sample_axes(data))

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def ensemble_std(self, *args, times=None, **kwargs):
        '''Standard deviation across series at each sampling time'''
        data = self.get_data(*args, times=times, **kwargs)
        return np.std(data, axis=self._data_sample_axes(data))

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def ensemble_median(self, *args, times=None, **kwargs):
        '''Median across series at each sampling time'''
        data = self.get_data(*args, times=times, **kwargs)
        return np.median(data, axis=self._data_sample_axes(data))

    @modify_help(arg_str='[series_idx], [times=[from,to]]')
    def ensemble_rms(self, *args, times=None, **kwargs):
        '''Root-Mean-Square across series at each sampling time'''
        x = self.get_data(*args, times=times, **kwargs)
        return np.sqrt(np.mean(np.abs(x)**2, axis=1))

    @modify_help(call='power(from_freq=xx, to_freq=xx, [series_idx])')
    def power(self, *args, from_freq=None, to_freq=None,
              segment_factor=None, window='boxcar', **kwargs):
        '''Power Spectral Density across specified series'''

        if segment_factor is None:
            if self._segment_factor is None:
                self._segment_factor = 1.0
        else:
            if self._segment_factor != segment_factor:
                self._segment_factor = segment_factor
                self._power = None
        if self._window != window:
            self._power = None
            self._window = window
        if self._power is None:
            data = self._get_not_indexed_data()
            # Perform power on flattened data and restore the shape afterwards
            power = self._compute_power(self._data_flattened(data))
            self._power = power.reshape((power.shape[0],) + self._data_sample_shape(data))
        if from_freq is None:
            output = self._power
            self._last_cut_frequency = self._frequency
        else:
            ul = self._frequency <= to_freq
            dl = self._frequency >= from_freq
            ul |= np.isclose(self._frequency, to_freq)
            dl |= np.isclose(self._frequency, from_freq)
            lim = ul & dl
            self._last_cut_frequency = self._frequency[lim]
            output = self._power[lim]
        index = self.get_index_of(*args, **kwargs)
        return self._index_data(output, index)

    def _compute_power(self, data):

        if isinstance(data, NotAvailable):
            raise Exception('Cannot calculate power: data is not available')
        
        delta_time = self.delta_time
        if isinstance(delta_time, u.Quantity):
            value_hz = (1 / delta_time).to_value(u.Hz)
        else:
            value_hz = 1 / delta_time

        self._frequency, x = welch(data.T, value_hz,
                                   window=self._window,
                                   nperseg=data.shape[0] / self._segment_factor)
        df = np.diff(self._frequency)[0]
        return x.T * df


# ___oOo___



'''
Created on Oct 16, 2018

@author: lbusoni
'''
import numpy as np


def quadraticSum(arrayOfErrorsWithSign):
    ''' quadraticSum(arrayOfErrorsWithSign)
    Sum in quadrature of errors, considering sign

    5 = quadraticSum([3, 4])
    8 = quadraticSum(10, -6])

    Inputs can be numpy arrays
    '''
    total = 0.
    for err in arrayOfErrorsWithSign:
        total += np.sign(err) * err ** 2
    return np.sign(total) * np.sqrt(np.abs(total))



import astropy.units as u

from arte.dataelab.base_timeseries import BaseTimeSeries
from arte.time_series.indexer import Indexer
from arte.utils.unit_checker import separate_value_and_unit

signal_unit = u.def_unit('signal')



class BaseSlopes(BaseTimeSeries):
    '''Slopes recorded from a generic WFS
    
    Slopes are by default in xxxxyyyy order
    (all X slopes followed by all Y slopes).

    This class adds a default slope indexer accepting 'x' and 'y' arguments,

    Derived classes can change the order using a different indexer.
    '''

    def __init__(self, data, time_vector=None, astropy_unit=signal_unit, data_label='slopes', axes=None):
        super().__init__(data=data,
                         time_vector=time_vector,
                         astropy_unit=astropy_unit,
                         data_label=data_label,
                         axes=axes)
        self._indexer = Indexer()

    def get_index_of(self, *args, **kwargs):
        return self._indexer.sequential_xy(self.ensemble_size(), *args, **kwargs)

    def imshow(self, cut_wings=0):
        '''
        Display X and Y slope 2d images
        cut_wings=x means that colorbar is saturated for array values below x percentile
        and above 100-x percentile. Default is 0, i.e. all data are displayed; values below
        0 are forced to 0, values above 50 are set to 50.
        '''
        title = "left:" + self._data_label + "-X, right:" + self._data_label + "-Y"
        super().imshow(cut_wings=cut_wings, title=title)

    def vecshow(self):
        '''Display slopes as vector field'''
        sx2d = self.get_display('x').mean(axis=0)
        sy2d = self.get_display('y').mean(axis=0)
        sx2d, _ = separate_value_and_unit(sx2d)
        sy2d, _ = separate_value_and_unit(sy2d)

        import matplotlib.pyplot as plt
        plt.quiver(sx2d, sy2d)
        return plt



import abc
import functools
import numpy as np
from scipy.linalg import pinv
from arte.types.modal_coefficients import ModalCoefficients
from arte.types.mask import CircularMask, BaseMask
from arte.types.wavefront import Wavefront
from arte.types.slopes import Slopes


class BaseModalDecomposer(abc.ABC):
    """
    Generic modal decomposer class.
    """

    def __init__(self, n_modes=None):
        self.nModes = n_modes
        self.lastModesGenerator = None
        self.lastIM = None
        self.lastRank = None
        self.lastReconstructor = None
        self.lastMask = None

    @abc.abstractmethod
    def generator(self, nModes, circular_mask, user_mask, **kwargs):
        '''Override to return a modal generator instance'''

    def _numpy2coefficients(self, coeff_array):
        '''Override to convert the resulting numpy array
        to a specific return type, if needed'''
        return ModalCoefficients(coeff_array)

    def getLastRank(self):
        return self.lastRank

    def wfIm(self, modes_generator, modesIdx, user_mask, dtype):
        wf = modes_generator.getModesDict(modesIdx)
        nslopes = user_mask.as_masked_array().compressed().size
        im = np.zeros((len(modesIdx), nslopes), dtype=dtype)
        for i, idx in enumerate(modesIdx):
            wf_masked = np.ma.masked_array(wf[idx].data, mask=user_mask.mask())
            im[i, :] = wf_masked.compressed()
        return im
    
    def slopesIm(self, modes_generator, modesIdx, user_mask, dtype):

        if not hasattr(modes_generator, 'getDerivativeXDict') or \
           not hasattr(modes_generator, 'getDerivativeYDict'):
            raise NotImplementedError(f'Modes generator of type {modes_generator.__class__.__name__}'
                                      ' does not define methods to get derivatives')

        dx = modes_generator.getDerivativeXDict(modesIdx)
        dy = modes_generator.getDerivativeYDict(modesIdx)
        nslopes = user_mask.as_masked_array().compressed().size
        im = np.zeros((len(modesIdx), 2 * nslopes), dtype=dtype)

        for i, idx in enumerate(modesIdx):
            dx_masked = np.ma.masked_array(dx[idx].data, mask=user_mask.mask())
            dy_masked = np.ma.masked_array(dy[idx].data, mask=user_mask.mask())
            im[i, :] = np.hstack(
                (dx_masked.compressed(), dy_masked.compressed()))
        return im

    def _syntheticReconstructor(self, im_func, nModes,
                                circular_mask, user_mask=None,
                                dtype=float, start_mode=None,
                                return_rank=False, **kwargs):
        '''Generates a synthetic reconstructor given a certain IM function'''
        if user_mask is None:
            user_mask = circular_mask

        atol = kwargs.pop('atol', None)
        rtol = kwargs.pop('rtol', None)
        check_finite = kwargs.pop('check_finite', True)

        self._assert_types(circular_mask, user_mask)

        modes_generator = self.generator(nModes, circular_mask, user_mask, **kwargs)
        if start_mode is None:
            first_mode = modes_generator.first_mode()
        else:
            first_mode = start_mode

        self.lastModesGenerator = modes_generator

        modesIdx = list(range(first_mode, first_mode + nModes))
        im = im_func(modes_generator, modesIdx, user_mask, dtype)
        self.lastIM = im
        return pinv(im, return_rank=return_rank, atol=atol, rtol=rtol, check_finite=check_finite)

    @functools.cache
    def cachedSyntheticReconstructorFromWavefront(self, nModes,
                                                  circular_mask, user_mask=None,
                                                  dtype=float, start_mode=None,
                                                  return_rank=False, **kwargs):
        '''Generates a synthetic reconstructor instance using the
           modal basis returned by self.generator()'''
        return self._syntheticReconstructor(self.wfIm, nModes, 
                                            circular_mask, user_mask,
                                            dtype, start_mode, return_rank, **kwargs)

    @functools.cache
    def cachedSyntheticReconstructorFromSlopes(self, nModes,
                                                  circular_mask, user_mask=None,
                                                  dtype=float, start_mode=None,
                                                  return_rank=False, **kwargs):
        '''Generates a synthetic reconstructor instance using the
           modal basis returned by self.generator()'''
        return self._syntheticReconstructor(self.slopesIm, nModes,
                                            circular_mask, user_mask,
                                            dtype, start_mode, return_rank, **kwargs)

    def measureModalCoefficientsFromWavefront(
        self, wavefront, circular_mask, user_mask, nModes=None, dtype=float, **kwargs
    ):
        self._assert_types(circular_mask, user_mask, wavefront=wavefront)
        if nModes is None:
            nModes = self.nModes

        reconstructor, rank = self.cachedSyntheticReconstructorFromWavefront(
            nModes, circular_mask, user_mask, dtype=dtype, return_rank=True, **kwargs
        )

        wavefrontInMaskVector = np.ma.masked_array(
            wavefront.toNumpyArray(), user_mask.mask()
        ).compressed()
        wavefrontInMaskVectorNoPiston = (
            wavefrontInMaskVector - wavefrontInMaskVector.mean()
        )
        result = self._numpy2coefficients(
            np.dot(wavefrontInMaskVectorNoPiston, reconstructor)
        )
        # Remember last used values
        self.lastRank = rank
        self.lastMask = user_mask
        self.lastReconstructor = reconstructor
        return result

    def measureModalCoefficientsFromSlopes(self, slopes, circular_mask,
                                           user_mask=None, nModes=None, dtype=float):

        self._assert_types(circular_mask, user_mask, slopes=slopes)
        if nModes is None:
            nModes = self.nModes
        if user_mask is None:
            user_mask = circular_mask

        reconstructor, rank = self.cachedSyntheticReconstructorFromSlopes(
            nModes, circular_mask, user_mask, dtype=dtype, return_rank=True)

        slopesInMaskVector = np.hstack(
            (np.ma.masked_array(slopes.mapX(), user_mask.mask()).compressed(),
             np.ma.masked_array(slopes.mapY(), user_mask.mask()).compressed())
        )

        result = self._numpy2coefficients(np.dot(slopesInMaskVector, reconstructor))
        # Remember last used values
        self.lastRank = rank
        self.lastMask = user_mask
        self.lastReconstructor = reconstructor
        return result

    def _assert_types(self, circular_mask, user_mask=None, wavefront=None, slopes=None):
        '''
        Make sure that:
         1) circular_mask is of type CircularMask
         2) user_mask, if specified, is of type BaseMask
         3) user_mask, if specified, is fully contained into circular_mask
         4) wavefront, if specified, is of type Wavefront
         5) slopes, if specified, is of type Slopes

        Raise an AssertionError if not.
        '''
        if user_mask is None:
            user_mask = circular_mask
        assert isinstance(circular_mask, CircularMask), \
            'circular_mask argument must be of type CircularMask, instead is %s' % \
            circular_mask.__class__.__name__
        assert isinstance(user_mask, BaseMask), \
            'user_mask argument must be of type BaseMask, instead is %s' % \
            user_mask.__class__.__name__

        if not np.all(
                circular_mask.as_masked_array() * user_mask.as_masked_array()
                == user_mask.as_masked_array()):
            raise ValueError(
                'User mask must be fully contained in circular mask')

        if wavefront:
            assert isinstance(wavefront, Wavefront), (
                "wavefront argument must be of type Wavefront, instead is %s"
                % wavefront.__class__.__name__
            )
        if slopes:
            assert isinstance(slopes, Slopes), (
                "slopes argument must be of type Slopes, instead is %s"
                % slopes.__class__.__name__
            )



# -*- coding: utf-8 -*-
#########################################################
#
# who       when        what
# --------  ----------  ---------------------------------
# apuglisi  2019-09-28  Created
#
#########################################################


def multiton(cls):
    '''
    Multiton decorator

    Decorator that returns the same instance of a class
    every time it is instantiated with the same parameters.

    All parameters must be able to be passed to str() in order
    to build an hashable key.
    As a side effect, the class name becomes a function
    that returns an instance, rather than a class type instance.
    '''
    instances = {}

    def getinstance(*args):
        key = '.'.join(map(str, args))
        if key not in instances:
            instances[key] = cls(*(args[1:]))
        return instances[key]

    return getinstance


def multiton_id(cls):
    '''
    Multiton decorator for mutable types

    Decorator that returns the same instance of a class
    every time it is instantiated with the same parameters.

    Similar to "multiton", but uses the id of each argument
    to build an hashable key. This allows to pass things
    like dictionaries that will be recognized as identical even
    if their contents change, but risks not recognizing identical
    values of strings and numbers.
    As a side effect, the class name becomes a function
    that returns an instance, rather than a class type instance.
    '''
    instances = {}

    def getinstance(*args):
        ids = [str(id(x)) for x in args]
        key = '.'.join(ids)
        if key not in instances:
            instances[key] = cls(*(args[1:]))
        return instances[key]

    return getinstance



import datetime


class Timestamp(object):

    def __init__(self):
        self._now = datetime.datetime.now()

    def asNowString(self):
        return self._now.strftime("%Y%m%d_%H%M%S")

    def asTodayString(self):
        return self._now.strftime("%Y%m%d")

    @staticmethod
    def now():
        return Timestamp().asNowString()

    @staticmethod
    def today():
        return Timestamp().asTodayString()

    @staticmethod
    def nowUSec():
        ss = datetime.datetime.now()
        return ss.strftime('%Y%m%d_%H%M%S')

    def __str__(self):
        return self.asNowString()



# -*- coding: utf-8 -*-
#########################################################
#
# who       when        what
# --------  ----------  ---------------------------------
# apuglisi  2020-02-01  Created
#
#########################################################

import numpy as np
from collections import namedtuple
import astropy.units as u

from arte.utils.help import add_help
from arte.utils.unit_checker import unit_check


@add_help
class MagEstimator():
    '''
    Estimates magnitude from detector counts

    Parameters
    ----------
    total_adus: int * u.adu
        total ADUs recorded by the detector in one exposure
    telescope: class Telescope or equivalent
        a class that defines an area() method with a result in u.m**2
    detector_bandname: str
        one of the allowed bandnames (use MagEstimator.bandnames() to list)
    detector_bandwidth: float * u.nm, optional
        bandwidth, defaults to 300 nm
    detector_gain: int, optional
        detector EM gain, defaults to 1
    detector_adu_e_ratio: float * u.electron/u.adu
        detector ADU/e- conversion factor, defaults to 1.0
    detector_nsubaps: int, optional
        number of subapertures, defaults to 1
    detector_freq: float * u.Hz, optional
        detector frequency, defaults to 1.0 Hz
    wfs_transmission: float, optional
        overall wfs transmission from telescope aperture to detector,
        defaults to 1
    detector_qe: float * u.electron/u.ph, optional
        average QE over the considered band,
        defaults to 1.0

    Raises
    ------
    TypeError
        if the wrong astropy units are passed in the initializer

    Notes
    -----
    From https://www.eso.org/observing/etc/doc/skycalc/helpskycalc.html#mags ::

        (flux for zero mag in Vega system)
        Band   Photometric zeropoint in ergs/s/cm2/A
         U     U0 = 4.18023e-09
         B     B0 = 6.60085e-09
         V     V0 = 3.60994e-09
         R     R0 = 2.28665e-09
         I     I0 = 1.22603e-09
         Z     Z0 = 7.76068e-10
         Y     Y0 = 5.973e-10
         J     J0 = 3.12e-10
         H     H0 = 1.14e-10
         K     K0 = 3.94e-11
         L     L0 = 4.83e-12
         M     M0 = 2.04e-12
         N     N0 = 1.23e-13
         Q     Q0 = 6.8e-15

    '''

    _Band = namedtuple('Band', 'wl ergs')
    _bands = {
        'R': _Band(650e-9 * u.m, 2.28665e-9 * u.erg / u.s / u.cm ** 2 / u.angstrom),
        'I': _Band(900e-9 * u.m, 1.22603e-9 * u.erg / u.s / u.cm ** 2 / u.angstrom),
        'V': _Band(532e-9 * u.m, 3.60994e-9 * u.erg / u.s / u.cm ** 2 / u.angstrom),
        }

    @unit_check
    def __init__(self, telescope,
                       detector_bandname,
                       total_adus=1 * u.adu,
                       detector_bandwidth=300 * u.nm,
                       detector_freq=1.0 * u.Hz,
                       detector_gain=1,
                       detector_adu_e_ratio=1.0 * u.electron / u.adu,
                       detector_nsubaps=1,
                       wfs_transmission=1.0,
                       detector_qe=1.0 * u.electron / u.ph):

        if detector_bandname not in self._bands:
            raise ValueError('Unsupported band: ' + str(detector_bandname))

        self._total_adus = total_adus
        self._detector_gain = detector_gain
        self._detector_adu_e_ratio = detector_adu_e_ratio
        self._detector_nsubaps = detector_nsubaps
        self._detector_freq = detector_freq
        self._wfs_transmission = wfs_transmission
        self._bandname = detector_bandname
        self._bandwidth = detector_bandwidth
        self._detector_qe = detector_qe
        self._telescope = telescope

    def bandname(self):
        '''Name of the band used for magnitude estimation'''
        return self._bandname

    def bandnames(self):
        '''List of supported bandnames'''
        return self._bands.keys()

    def flux_zero(self):
        '''Zero point in photons/sec'''

        wl, ergs = self._bands[self._bandname]

        photons = self._ergs_to_photons(ergs, wl)

        return (photons * self._telescope.area() * self._bandwidth).to(u.ph / u.s)

    def _ergs_to_photons(self, ergs, wl):
        '''
        Converts from ergs to number of photons at a given wavelength

        Uses the ergs/cm^2/s/A to photons/cm^2/s/A constants from:
        https://hea-www.harvard.edu/~pgreen/figs/Conversions.pdf
        http://www.stsci.edu/~strolger/docs/UNITS.txt
        '''
        return ergs * 5.0341125e7 * wl.to(u.angstrom).value * (u.ph / u.erg)

    def photons_per_subap_per_frame(self):
        '''Photons/subap/frame detected by sensor'''
        gain = self._detector_gain
        nsubaps = self._detector_nsubaps
        adu_e_ratio = self._detector_adu_e_ratio
        qe = self._detector_qe

        return self._total_adus * adu_e_ratio / gain / nsubaps / qe

    def photons_per_second(self):
        '''Photons/sec detected by sensor'''
        ph_subap_frame = self.photons_per_subap_per_frame()

        freq = self._detector_freq.to('1/s')
        nsubaps = self._detector_nsubaps
        transmission = self._wfs_transmission

        return ph_subap_frame * freq * nsubaps / transmission

    def mag(self):
        '''Estimated magnitude'''
        flux = self.photons_per_second()
        zero_flux = self.flux_zero()

        return -2.5 * np.log10(flux / zero_flux)




import numpy as np
import astropy.units as u

from matplotlib import pyplot
from matplotlib.axes import Axes

from arte.utils.not_available import NotAvailable
from arte.utils.unit_checker import make_sure_its_a, separate_value_and_unit


def modalplot(residual_modes_vector, pol_modes_vector, unit=None,
            overplot=False, plot_to=None,
            title=None, xlabel='Mode index', ylabel='wavefront rms', add_unit_to_ylabel=True,
            res_label='residual modes', res_color='red',
            pol_label='POL modes', pol_color='black'):
    '''
    Example to apply an optical gain of 2:

    modalplot(self.residual_modes.time_std() * 2, 
              self.pol_modes.time_std() * 2,
              )
    '''

    if plot_to is not None:
        plt = plot_to
    else:
        plt = pyplot

    res_data = residual_modes_vector
    pol_data = pol_modes_vector

    if not overplot:
        plt.cla()
        plt.clf()
        if title is None:
            raise ValueError('Title must be set when not overplotting')

    # Make sure that the two units match if at all possible,
    # taking care of leaving NotAvailable values untouched.
    res_valid = not isinstance(res_data, NotAvailable)
    pol_valid = not isinstance(pol_data, NotAvailable)
    res_quantity = isinstance(res_data, u.Quantity)
    pol_quantity = isinstance(pol_data, u.Quantity)
    if unit is not None:
        res_data = make_sure_its_a(unit, res_data)
        pol_data = make_sure_its_a(unit, pol_data)
    elif res_valid and pol_valid and (res_quantity or pol_quantity):
        pol_data = make_sure_its_a(res_data.unit, pol_data)

    res_value, res_unit = separate_value_and_unit(res_data)
    pol_value, pol_unit = separate_value_and_unit(pol_data)

    # Select what to display on the Y axis.
    plot_unit = ''
    if res_quantity:
        plot_unit = res_unit
    elif pol_quantity:
        plot_unit = pol_unit
    if not isinstance(plot_unit, u.UnitBase):
        plot_unit = 'a.u.'
    
    if len(res_data) > 0:
        plt.plot(np.arange(len(res_data)), res_value,
                    color=res_color, label=res_label)
    if len(pol_data) > 0:
        plt.plot(np.arange(len(pol_data)), pol_value,
                    color=pol_color, label=pol_label)

    maxlen = max(len(res_data), len(pol_data))

    if add_unit_to_ylabel:
        ystr = f'{ylabel} [{plot_unit}]'
    else:
        ystr = ylabel

    if not overplot:
        if isinstance(plt, Axes):
            plt.set(xlabel=xlabel, ylabel=ystr)
            if title is not None:
                plt.set(title=title)
        else:
            plt.loglog()
            plt.xscale('symlog')    # Allow 0 index in log plots
#            plt.yscale('symlog')
            plt.xlim([0, maxlen])

            plt.xlabel(xlabel)
            plt.ylabel(ystr)
            if title is not None:
                plt.title(title)
            if not (res_label is None and pol_label is None):
                plt.legend()
    return plt




import time


class TimeKeeper:
    '''
    Class to give information about iterative loops.

    Every time the inc() method is called, an internal counter is
    incremented, and other members keep track of the increment rate
    and average interval between increments.

    The inc() method returns True if at least *interval* seconds
    have passed since the class was initialized or since the last time
    True was returned, otherwise it will return False. When True
    is returned, the internal counter and rate calculations are resetted.

    Example:

    timekeep = TimeKeeper()
    while True:
        time.sleep(0.1)
        if timekeep.inc():
            print('%d iterations (%5.2f Hz) - average time: %5.2f ms' %
                      (timekeep.count, timekeep.rate, timekeep.ms))
    '''

    def __init__(self, interval=1.0):
        self.interval = interval
        self.t0  = time.time()
        self.cnt = 0

    def inc(self):
        self.cnt += 1
        t1 = time.time()
        if (t1 - self.t0) >= self.interval:
            tdiff = t1 - self.t0
            self.rate = self.cnt / tdiff
            self.count = self.cnt
            self.ms = 1.0 / self.rate * 1e3
            self.cnt = 0
            self.t0 = t1
            return True
        return False


class CyclePrinter():
    '''
    Class to give information about iterative loops

    Initialize with a name (that will appear as the first word in the
    output message), an optional minimum period between messages
    in seconds (defualt 1.0), and an optional function that will be
    called with a single string argument (defaults to the system
    "print" function), and a format for the number of seconds.

    Call the 'cycle' function at each iteration. If at least "period"
    second have passed from the last call, the "log_func" function
    will be called with a descriptive message as a string argument.
    '''

    def __init__(self, name, period=1.0, log_func=print, fmt='%5.2f'):
        self.name = name
        self.period = period
        self.cycle_counter = 0
        self.start = None
        self.prev_cycle = None
        self.log_func = log_func
        self.fmt = fmt

    def cycle(self):
        self.cycle_counter += 1
        now = time.time()
        if self.start is None:
            self.start = now
            self.prev_cycle = now
            return

        if now - self.prev_cycle > self.period:
            self.elapsed_time = now - self.prev_cycle
            self.log_func(self._msg())
            self.prev_cycle = now
            self.cycle_counter = 0

    def _msg(self):
        msg = ('%s: %d cycles in ' + self.fmt + ' seconds') % \
               (self.name,
                self.cycle_counter,
                self.elapsed_time)
        return msg


class PercentPrinter(CyclePrinter):
    '''
    Small class to give information about iterative loops

    Specializaton of CyclePrinter for processes that go from 0% to 100%.
    Initialize with the same arguments as CyclePrinter and an additional
    "total" argument that represents the number of times the "cycle"
    function will be called. "total" defaults to 100.
    '''

    def __init__(self, name, period=1.0, log_func=print,
                       fmt='%5.2f', total=100):

        CyclePrinter.__init__(self, name, period, log_func, fmt)
        self.total = total

    def _msg(self):

        percent = int(self.counter * 100.0 / self.total)
        msg = ('%s: %d done (%d out of %d) in ' + self.fmt + ' seconds') % \
                (self.name,
                 percent,
                 self.counter,
                 self.total,
                 self.elapsed_time)
        return msg


class SpeedPrinter(CyclePrinter):
    '''
    Small class to give information about iterative loops

    Similar to CyclePrinter, but prints the loop speed
    instead of the number of cycles
    '''

    def __init__(self, name, period=1.0, log_func=print, fmt='%5.2f'):
        CyclePrinter.__init__(self, name, period, log_func, fmt)
        self.fmt = fmt

    def _msg(self):

        speed = self.cycle_counter / self.elapsed_time
        msg = ('%s: ' + self.fmt + ' iteration/sec') % (self.name, speed)
        return msg



import numpy as np
from arte.utils.zernike_generator import ZernikeGenerator
from arte.atmo.utils import getFullKolmogorovCovarianceMatrix
from arte.utils.karhunen_loeve_generator import KarhunenLoeveGenerator as  KLGenerator
from arte.utils.base_modal_decomposer import BaseModalDecomposer


class KarhunenLoeveModalDecomposer(BaseModalDecomposer):
    """
    This class decomposes a wavefront into a set of Karhunen Loeve modal coefficients.

    Parameters
    ----------
    n_modes: int
        Number of modes to decompose the wavefront into.
    """

    def __init__(self, n_modes=None):
        super().__init__(n_modes)

    def generator(self, nModes, circular_mask, user_mask, **kwargs):
        zz = ZernikeGenerator(circular_mask)
        zbase = np.rollaxis(
            np.ma.masked_array([zz.getZernike(n) for n in range(2, self.nModes + 2)]),
            0,
            3,
        )
        kl = KLGenerator(circular_mask, getFullKolmogorovCovarianceMatrix(self.nModes))
        kl.generateFromBase(zbase)
        return kl





import numpy as np
import astropy.units as u

from arte.utils.help import modify_help
from arte.time_series.time_series import TimeSeries

class MultiTimeSeries(TimeSeries):
    '''
    Join multiple TimeSeries objects with incompatible time sampling.
    
    After being initialized with a list of
    :class:`~arte.time_series.time_series.TimeSeries` objects, which can
    have different ensemble sizes and length, it behaves as if all the data
    was part of a single ensemble. Operations that make calculations
    time-wise (like 
    :meth:`~arte.time_series.time_series.TimeSeries.time_std()`) work in any
    case. The ensemble-wise ones
    (like :meth:`~arte.time_series.time_series.TimeSeries.ensemble_std()`)
    only work if deltaTime is the same
    across all series, and raise an Exception otherwise.
    '''
    def __init__(self, *args):
        super().__init__()

        self._series = []
        for v in args:
            self.add_series(v)

    def add_series(self, series):
        '''
        Adds a new series to this MultiTimeSeries instance
        
        Parameters
        ----------
        series: :class:`~arte.time_series.time_series.TimeSeries` or :class:`~arte.time_series.time_series.TimeSeriesWithInterpolation` instance
           the series to be added        
        '''
        self._series.append(series)

    def _get_not_indexed_data(self):
        return np.hstack([v._get_not_indexed_data()
                          for v in self._series])
        
    def is_homogeneous(self, *args, **kwargs):
        '''Returns True if all selected series have the same
        :py:attr:`~arte.time_series.time_series.TimeSeries.delta_time`
        '''
        dt = [x.value for x in self.delta_times(*args, **kwargs)]
        return len(set(dt)) == 1

    def _impersonateDeltaTime(self, *args, **kwargs):
        '''Assume the first delta time of the ones selected by the args'''
        dt = self.delta_times(*args, **kwargs)
        self.delta_time = dt[0]

    def ensemble_size(self):
        '''Returns the total ensemble size'''
        return sum([x.ensemble_size() for x in self._series])

    def delta_times(self, *args, **kwargs):
        '''Returns a vector of delta times'''

        # Known astropy bug (numpy < 1.17): units are lost when using hstack 
        # We remove them before stacking, and add them later
        dt = np.hstack( \
              [np.repeat(x.delta_time.to('s').value, x.ensemble_size())
               for x in self._series])

        dt = dt * u.s

        index = self.get_index_of(*args, **kwargs)
        if index is not None and len(index)>0:
            dt = dt[index]

        return dt

    @modify_help(arg_str='[time_idx]')
    def ensemble_average(self, *args, times=None, **kwargs):
        ''' Average across series at each sampling time '''
        if self.is_homogeneous(*args, **kwargs):
            self._impersonateDeltaTime(*args, **kwargs) 
            return super().ensemble_average(*args, times=times, **kwargs)
        else:
            raise Exception('Data series cannot be combined')

    @modify_help(arg_str='[time_idx]')
    def ensemble_std(self, *args, times=None, **kwargs):
        ''' Standard deviation across series at each sampling time '''
        if self.is_homogeneous(*args, **kwargs):
            self._impersonateDeltaTime(*args, **kwargs) 
            return super().ensemble_std(*args, times=times, **kwargs)
        else:
            raise Exception('Data series cannot be combined')

    @modify_help(arg_str='[time_idx]')
    def ensemble_median(self, *args, times=None, **kwargs):
        ''' Standard deviation across series at each sampling time '''
        if self.is_homogeneous(*args, **kwargs):
            self._impersonateDeltaTime(*args, **kwargs) 
            return super().ensemble_median(*args, times=times, **kwargs)
        else:
            raise Exception('Data series cannot be combined')

# ___oOo___


# -*- coding: utf-8 -*-
##########################################
# who       when        what
# --------  ----------  ------------------
# apuglisi  2019-09-28  Created
#
##########################################


from arte.math.factors import lcm
from arte.utils.rebin import rebin


def toccd(a, newshape, set_total=None):
    '''
    Clone of oaalib's toccd() function, using least common multiple
    to rebin an array similar to opencv's INTER_AREA interpolation.
    '''
    if a.shape == newshape:
        return a

    if len(a.shape) != 2:
        raise ValueError('Input array shape is %s instead of 2d, cannot continue:' % str(a.shape))

    if len(newshape) != 2:
        raise ValueError('Output shape is %s instead of 2d, cannot continue' % str(newshape))

    if set_total is None:
        set_total = a.sum()

    mcmx = lcm(a.shape[0], newshape[0])
    mcmy = lcm(a.shape[1], newshape[1])

    temp = rebin(a, (mcmx, a.shape[1]), sample=True)
    temp = rebin(temp, (newshape[0], a.shape[1]))
    temp = rebin(temp, (newshape[0], mcmy), sample=True)
    rebinned = rebin(temp, newshape)

    return rebinned / rebinned.sum() * set_total





import logging

def if_(condition, warning=None):
    '''
    Decorator to turn a function into a NO-OP
    if the condition is not met.
    Source: https://stackoverflow.com/questions/17946024/deactivate-function-with-decorator

    Example:
    @if_(global_enable)
    def do_something():
        ...

    '''
    def noop_decorator(func):
        return func  # pass through

    def neutered_function(func):
        def neutered(*args, **kw):
            if warning:
                logging.warn(warning)
            return None
        return neutered

    return noop_decorator if condition else neutered_function




'''
@author: giuliacarla
'''

import numpy as np
try:
    import cupy as cp
except ImportError:
    print("Can't import cupy")
    cp = None
import astropy.units as u
from arte.utils import math
import logging
from arte.utils.zernike_generator import ZernikeGenerator
from arte.atmo import von_karman_psd
# from arte.utils.decorator import cacheResult


class VonKarmanSpatioTemporalCovariance():
    """
    Covariance of Von Karman atmospheric turbulence

    This class computes the covariance and its Fourier transform, the Cross
    Power Spectral Density (CPSD), between Zernike coefficients describing
    the turbulence induced phase aberrations of two sources seen by two
    different circular apertures. The CPSD of the phase is also computed.


    References
    ----------
        Plantet et al. (2020) - "Spatio-temporal statistics of the turbulent
        Zernike coefficients and piston-removed phase from two distinct beams".

        Plantet et al. (2018) - "LO WFS of MAORY: performance and sky coverage
        assessment."

        Whiteley et al. (1998) - "Temporal properties of the Zernike expansion
        coefficients of turbulence-induced phase aberrations for aperture
        and source motion".


    Parameters
    ----------
    source1: `~arte.types.guide_source.GuideSource`
        Geometry of the first source.
        We consider rho as the angle in arcsec wrt the z-axis and
        theta as the angle in degrees wrt the x-axis.
        (e.g. source1 = arte.types.guide_source.GuideSource((1,90), 9e3)

    source2: `~arte.types.guide_source.GuideSource`
        Geometry of the second source. Same conventions as source1.

    aperture1: `~arte.types.aperture.CircularOpticalAperture`
        Geometry of the first optical aperture.
        (e.g. aperture1 = arte.types.aperture.CircularOpticalAperture(
        10, (0, 0, 0)))

    aperture2: `~arte.types.aperture.CircularOpticalAperture`
        Geometry of the second optical aperture.

    cn2_profile: `~arte.atmo.cn2_profile`
        Cn2 profile.
        (e.g. cn2_eso = arte.atmo.cn2_profile.EsoEltProfiles.Q1()
        e.g. cn2_invented = arte.atmo.cn2_profile.Cn2Profile.from_r0s(
        [0.16], [25], [10e3], [0.1], [0]))

    spat_freqs: `numpy.ndarray`
        Range of spatial frequencies that are used in Zernike covariance,
        Zernike CPSD and phase CPSD computation.

    """

    def __init__(self,
                 source1,
                 source2,
                 aperture1,
                 aperture2,
                 cn2_profile,
                 spat_freqs,
                 logger=logging.getLogger('VK_COVARIANCE')):
        self.setSource1(source1)
        self.setSource2(source2)
        self.setAperture1(aperture1)
        self.setAperture2(aperture2)
        self.setCn2Profile(cn2_profile)
        self.setSpatialFrequencies(spat_freqs)
        self._xp = np
        self._logger = logger

    def useGPU(self):
        self._xp = cp

    def setCn2Profile(self, cn2_profile):
        self._resetCachedResults()
        self._cn2 = cn2_profile
        self._layersDistance = self._quantitiesToValue(
            cn2_profile.layers_distance())
        self._r0s = self._quantitiesToValue(cn2_profile.r0s())
        self._outerScale = self._quantitiesToValue(cn2_profile.outer_scale())
        self._numberOfLayers = self._quantitiesToValue(
            cn2_profile.number_of_layers())
        self._windSpeed = self._quantitiesToValue(cn2_profile.wind_speed())
        self._windDirection = self._quantitiesToValue(
            cn2_profile.wind_direction())

    def setSource1(self, s1):
        self._resetCachedResults()
        self._source1 = s1
        self._source1Coords = self._quantitiesToValue(
            s1.getSourcePolarCoords())

    def setSource2(self, s2):
        self._resetCachedResults()
        self._source2 = s2
        self._source2Coords = self._quantitiesToValue(
            s2.getSourcePolarCoords())

    def setAperture1(self, ap1):
        self._resetCachedResults()
        self._ap1 = ap1
        self._ap1Coords = self._quantitiesToValue(
            ap1.getCartesianCoords())
        self._ap1Radius = self._quantitiesToValue(
            ap1.getApertureRadius())

    def setAperture2(self, ap2):
        self._resetCachedResults()
        self._ap2 = ap2
        self._ap2Coords = self._quantitiesToValue(
            ap2.getCartesianCoords())
        self._ap2Radius = self._quantitiesToValue(
            ap2.getApertureRadius())

    def setSpatialFrequencies(self, freqs):
        self._resetCachedResults()
        self._spat_freqs = freqs

    def _resetCachedResults(self):
        # TODO: lb 200620. This is ugly, the decorator should offer
        # the method to reset itself. It could also mean that the
        # methods setSource, setAperture etc are not needed and one
        # should instantiate a new object instead of modifying it.
        # Also, note that we call this method to reset everything
        # every time something is modified, and this is
        # likely suboptimal
        self._initializeGeometry_cached_result = {}

    def source1(self):
        return self._source1

    def source2(self):
        return self._source2

    def aperture1(self):
        return self._ap1

    def aperture2(self):
        return self._ap2

    def cn2_profile(self):
        return self._cn2

    def spatial_frequencies(self):
        return self._spat_freqs

    def temporal_frequencies(self):
        return self._temporalFreqs

    def _quantitiesToValue(self, quantity):
        if type(quantity) == list:
            value = np.array([quantity[i].value for i in range(len(quantity))])
        else:
            value = quantity.value
        return value

    def _layerScalingFactor1(self, nLayer):
        a1 = self._layerScalingFactor(
            self._layersDistance[nLayer],
            self._source1Coords[2],
            self._ap1Coords[2])
        return a1

    def _layerScalingFactor2(self, nLayer):
        a2 = self._layerScalingFactor(
            self._layersDistance[nLayer],
            self._source2Coords[2],
            self._ap2Coords[2])
        return a2

    def _layerScalingFactor(self, zLayer, zSource, zAperture):
        scalFact = (zLayer - zAperture) / (zSource - zAperture)
        return scalFact

    def _getCleverVersorCoords(self, s_coord):
        return np.array([
            np.sin(np.deg2rad(s_coord[0] / 3600)) *
            np.cos(np.deg2rad(s_coord[1])),
            np.sin(np.deg2rad(s_coord[0] / 3600)) *
            np.sin(np.deg2rad(s_coord[1])),
            np.cos(np.deg2rad(s_coord[0] / 3600))])

    def _cleverVersor1Coords(self):
        return self._getCleverVersorCoords(
            self._source1Coords)

    def _cleverVersor2Coords(self):
        return self._getCleverVersorCoords(
            self._source2Coords)

    def _layerProjectedAperturesSeparation(self, nLayer):
        aCoord1 = self._ap1Coords
        aCoord2 = self._ap2Coords

        vCoord1 = self._cleverVersor1Coords()
        vCoord2 = self._cleverVersor2Coords()

        sep = aCoord2 - aCoord1 + (
            self._layersDistance[nLayer] - aCoord2[2]) * \
            vCoord2 / vCoord2[2] - \
            (self._layersDistance[nLayer] - aCoord1[2]) * \
            vCoord1 / vCoord1[2]
        return sep

    def _VonKarmanPSDOneLayer(self, nLayer, freqs):
        vk = von_karman_psd.VonKarmanPsd(
            self._r0s[nLayer], self._outerScale[nLayer])
        psd = vk.spatial_psd(freqs)
        return psd

#     @cacheResult
    def _initializeGeometry(self, nLayer, spat_freqs):
        a1 = self._layerScalingFactor1(nLayer)
        a2 = self._layerScalingFactor2(nLayer)
        sep = self._xp.array(self._layerProjectedAperturesSeparation(nLayer))

        if self._xp == cp:
            psd = self._VonKarmanPSDOneLayer(
                nLayer, self._xp.asnumpy(spat_freqs))
            psd = self._xp.array(psd)
        elif self._xp == np:
            psd = self._VonKarmanPSDOneLayer(nLayer, spat_freqs)
        sl = self._xp.linalg.norm(sep)
        thS = self._xp.arctan2(sep[1], sep[0])
        return a1, a2, sl, thS, psd

    def _initializeFunctionsForZernikeComputations(
            self, j, k, spat_freqs, a1, a2):
        nj, mj = ZernikeGenerator.degree(j)
        nk, mk = ZernikeGenerator.degree(k)

        deltaj = math.kroneckerDelta(0, mj)
        deltak = math.kroneckerDelta(0, mk)

        if self._xp == cp:
            r1 = cp.array(self._ap1Radius)
            r2 = cp.array(self._ap2Radius)
            b1 = math.besselFirstKindOnGPU(
                nj + 1,
                2 * self._xp.pi * spat_freqs * r1 * (1 - a1))
            b2 = math.besselFirstKindOnGPU(
                nk + 1,
                2 * self._xp.pi * spat_freqs * r2 * (1 - a2))
        elif self._xp == np:
            r1 = self._ap1Radius
            r2 = self._ap2Radius
            b1 = math.besselFirstKind(
                nj + 1,
                2 * self._xp.pi * spat_freqs * r1 * (1 - a1))
            b2 = math.besselFirstKind(
                nk + 1,
                2 * self._xp.pi * spat_freqs * r2 * (1 - a2))

        c0 = (-1) ** mk * self._xp.sqrt((nj + 1) * (nk + 1)) * 1j ** (
            nj + nk) * 2 ** (
            1 - 0.5 * (deltaj + deltak))
        c1 = 1. / (
            self._xp.pi * r1 * r2 * (1 - a1) * (1 - a2))

        return nj, mj, nk, mk, deltaj, deltak, b1, b2, c0, c1

    def _computeZernikeCovarianceOneLayer(self, j, k, nLayer):
        f = self._xp.array(self._spat_freqs)
        a1, a2, sl, thS, psd = \
            self._initializeGeometry(nLayer, f)
        _, mj, _, mk, deltaj, deltak, b1, b2, c0, c1 = \
            self._initializeFunctionsForZernikeComputations(
                j, k, f, a1, a2)

        if self._xp == cp:
            b3 = math.besselFirstKindOnGPU(
                mj + mk,
                sl * 2 * self._xp.pi * f)
            b4 = math.besselFirstKindOnGPU(
                np.abs(mj - mk),
                sl * 2 * self._xp.pi * f)
        elif self._xp == np:
            b3 = math.besselFirstKind(
                mj + mk,
                sl * 2 * self._xp.pi * f)
            b4 = math.besselFirstKind(
                np.abs(mj - mk),
                sl * 2 * self._xp.pi * f)
        c2 = self._xp.pi / 4 * ((1 - deltaj) * ((-1) ** j - 1) +
                                (1 - deltak) * ((-1) ** k - 1))
        c3 = self._xp.pi / 4 * ((1 - deltaj) * ((-1) ** j - 1) -
                                (1 - deltak) * ((-1) ** k - 1))

        zernikeIntegrand = c0 * c1 * \
            psd / f * b1 * b2 * \
            (self._xp.cos((mj + mk) * thS + c2) *
             1j ** (3 * (mj + mk)) *
             b3 +
             np.cos((mj - mk) * thS + c3) *
             1j ** (3 * self._xp.abs(mj - mk)) *
             b4)
        if self._xp == cp:
            zernikeIntegrand = self._xp.asnumpy(zernikeIntegrand)
            f = self._xp.asnumpy(f)
        zernikeCovOneLayer = np.trapz(zernikeIntegrand, f)
        return zernikeCovOneLayer

    def _computeZernikeCovarianceAllLayers(self, j, k):
        zernikeCovAllLayers = np.array([
            self._computeZernikeCovarianceOneLayer(j, k, nLayer) for nLayer
            in range(self._numberOfLayers)])
        return zernikeCovAllLayers.sum()

    def _initializeFunctionsForPhaseComputations(self, f, a1, a2):
        if self._xp == cp:
            r1 = cp.array(self._ap1Radius)
            r2 = cp.array(self._ap2Radius)
            k = (1 - a2) * r2 / ((1 - a1) * r1)
            arg1 = self._xp.pi * r1 * (1 - a1)
            arg2 = self._xp.pi * r2 * (1 - a2)

            b0 = math.besselFirstKindOnGPU(1, 2 * arg1 * (k - 1) * f)
#             def b0(f): return math.besselFirstKindOnGPU(
#                 1, 2 * arg1 * (k - 1) * f)

            b1 = math.besselFirstKindOnGPU(1, 2 * arg1 * f)
#             def b1(f): return math.besselFirstKindOnGPU(
#                 1, 2 * arg1 * f)

            b2 = math.besselFirstKindOnGPU(1, 2 * arg2 * f)
#             def b2(f): return math.besselFirstKindOnGPU(
#                 1, 2 * arg2 * f)
        elif self._xp == np:
            k = (1 - a2) * self._ap2Radius / ((1 - a1) * self._ap1Radius)
            arg1 = np.pi * self._ap1Radius * (1 - a1)
            arg2 = np.pi * self._ap2Radius * (1 - a2)
            b0 = math.besselFirstKind(1, 2 * arg1 * (k - 1) * f)
#             def b0(f): return math.besselFirstKind(
#                 1, 2 * arg1 * (k - 1) * f)
            b1 = math.besselFirstKind(1, 2 * arg1 * f)
#             def b1(f): return math.besselFirstKind(
#                 1, 2 * arg1 * f)
            b2 = math.besselFirstKind(1, 2 * arg2 * f)
#             def b2(f): return math.besselFirstKind(
#                 1, 2 * arg2 * f)

        if k == 1:
            #             self._b1Phase = lambda f: f**0
            self._b1Phase = 1
        else:
            #             self._b1Phase = lambda f: b0(
            #                 f) / (arg1 * (k - 1) * f)
            self._b1Phase = b0 / (arg1 * (k - 1) * f)

        if a1 == 1:
            #             self._b2Phase = lambda f: f**0
            self._b2Phase = 1
        else:
            #             self._b2Phase = lambda f: b1(
            #                 f) / (arg1 * f)
            self._b2Phase = b1 / (arg1 * f)

        if a2 == 1:
            #             self._b3Phase = lambda f: f**0
            self._b3Phase = 1
        else:
            #             self._b3Phase = lambda f: b2(
            #                 f) / (arg2 * f)
            self._b3Phase = b2 / (arg2 * f)

    def _computePhaseCovarianceOneLayer(self, nLayer):
        a1, a2, sl, thS, psd = \
            self._initializeGeometry(nLayer, self._spat_freqs)
        self._initializeFunctionsForPhaseComputations(
            self._spat_freqs, a1, a2)
        b0Phase = math.besselFirstKind(0,
                                       2 * np.pi * self._spat_freqs * sl)

        phaseIntegrand = 2 * np.pi * self._spat_freqs * psd * b0Phase * (
            self._b1Phase - self._b2Phase * self._b3Phase)
        phaseCovOneLayer = np.trapz(phaseIntegrand, self._spat_freqs)
        return phaseCovOneLayer

    def _getZernikeCPSDAllTemporalFrequenciesOneLayer(self, j, k, nLayer,
                                                      temp_freqs):

        integrand, fPerp = self.integrandOfZernikeCPSD(j, k, nLayer,
                                                       temp_freqs)
        cpsdZernikeOneLayer = np.trapz(integrand, fPerp, axis=0)
        return cpsdZernikeOneLayer

    def _getZernikeCPSDAllLayers(self, j, k, temp_freqs):
        cpsd = np.array([
            self._getZernikeCPSDAllTemporalFrequenciesOneLayer(
                j, k, nLayer, temp_freqs) for nLayer
            in range(self._numberOfLayers)])
        cpsdTotal = cpsd.sum(axis=0)
        return cpsdTotal

    def _getGeneralZernikeCPSDAllTemporalFrequenciesOneLayer(self, j, k,
                                                             nLayer,
                                                             temp_freqs):

        integrand, fPerp = self.integrandOfGeneralZernikeCPSD(j, k, nLayer,
                                                              temp_freqs)
        cpsdGeneralZernikeOneLayer = np.trapz(integrand, fPerp, axis=0)
        return cpsdGeneralZernikeOneLayer

    def _getGeneralZernikeCPSDAllLayers(self, j, k, temp_freqs):
        cpsd = np.array([
            self._getGeneralZernikeCPSDAllTemporalFrequenciesOneLayer(
                j, k, nLayer, temp_freqs) for nLayer
            in range(self._numberOfLayers)])
        cpsdTotal = cpsd.sum(axis=0)
        return cpsdTotal

    def _getPhaseCPSDAllTemporalFrequenciesOneLayer(self, nLayer, t_freqs):
        func, fPerp = self.integrandOfPhaseCPSD(nLayer, t_freqs)
        cpsdPhaseOneLayer = np.trapz(func, fPerp, axis=0)
        return cpsdPhaseOneLayer

    def _getGeneralPhaseCPSDAllTemporalFrequenciesOneLayer(self, nLayer,
                                                           t_freqs):
        func, fPerp = self.integrandOfGeneralPhaseCPSD(nLayer, t_freqs)
        genCpsdPhaseOneLayer = np.trapz(func, fPerp, axis=0)
#         inte = self.integrandOfGeneralPhaseCPSD(nLayer, t_freqs)(
#             self._spat_freqs)
#         genCpsdPhaseOneLayer, _ = integrate.quad_vec(inte,
#                                                      self._spat_freqs[0],
#                                                      self._spat_freqs[
#                                                          self._spat_freqs.shape[0]
#                                                          - 1])
#         genCpsdPhaseOneLayer = np.trapz(inte, self._spat_freqs, axis=0)
        return genCpsdPhaseOneLayer

    def _getPhaseCPSDAllLayers(self, temp_freqs):
        phaseCPSD = np.array([
            self._getPhaseCPSDAllTemporalFrequenciesOneLayer(
                nLayer, temp_freqs)
            for nLayer in range(self._numberOfLayers)])
        phaseCPSDTotal = phaseCPSD.sum(axis=0)
        return phaseCPSDTotal

    def _getGeneralPhaseCPSDAllLayers(self, temp_freqs):
        gPhaseCPSD = np.array([
            self._getGeneralPhaseCPSDAllTemporalFrequenciesOneLayer(
                nLayer, temp_freqs)
            for nLayer in range(self._numberOfLayers)])
        gPhaseCPSDTotal = gPhaseCPSD.sum(axis=0)
        return gPhaseCPSDTotal

    def integrandOfGeneralZernikeCPSD(self, j, k, nLayer, temp_freqs):
        if self._xp == cp:
            vl = self._xp.array(self._windSpeed[nLayer])
            vdir = self._xp.array(self._windDirection[nLayer])
            tf, sp_perp = self._xp.meshgrid(
                self._xp.array(temp_freqs), self._xp.array(self._spat_freqs))
        elif self._xp == np:
            vl = self._windSpeed[nLayer]
            vdir = self._windDirection[nLayer]
            tf, sp_perp = self._xp.meshgrid(temp_freqs, self._spat_freqs)

        f = self._xp.sqrt(sp_perp ** 2 + (tf / vl) ** 2)
        a1, a2, sl, thS, psd = self._initializeGeometry(nLayer, f)
        nj, mj, nk, mk, deltaj, deltak, b1, b2, c0, c1 = \
            self._initializeFunctionsForZernikeComputations(
                j, k, f, a1, a2)

        thWind = self._xp.deg2rad(vdir)
        th0 = self._xp.arccos(-tf / (f * vl))
        th1 = th0 + thWind
        th2 = -th0 + thWind

        c4 = self._xp.pi / 4 * (1 - deltaj) * ((-1) ** j - 1)
        c5 = self._xp.pi / 4 * (1 - deltak) * ((-1) ** k - 1)

        integFunc = 2 * c0 * c1 * 1j ** (0.5 * ((-1) ** (
            nj + nk) - 1)) / (vl * self._xp.pi) * \
            psd / f ** 2 * b1 * b2 * \
            (self._xp.cos(2 * self._xp.pi * f * sl * self._xp.cos(th1 - thS) +
                          self._xp.pi / 4 * ((-1) ** (nj + nk) - 1)) *
             self._xp.cos(mj * th1 + c4) *
             self._xp.cos(mk * th1 + c5) +
             self._xp.cos(2 * self._xp.pi * f * sl * self._xp.cos(th2 - thS) +
                          self._xp.pi / 4 * ((-1) ** (nj + nk) - 1)) *
             self._xp.cos(mj * th2 + c4) *
             self._xp.cos(mk * th2 + c5))

        if self._xp == cp:
            return self._xp.asnumpy(integFunc), self._xp.asnumpy(sp_perp)
        elif self._xp == np:
            return integFunc, sp_perp

    def integrandOfZernikeCPSD(self, j, k, nLayer, temp_freqs):
        if self._xp == cp:
            vl = self._xp.array(self._windSpeed[nLayer])
            vdir = self._xp.array(self._windDirection[nLayer])
            tf, sp_perp = self._xp.meshgrid(
                self._xp.array(temp_freqs), self._xp.array(self._spat_freqs))
        elif self._xp == np:
            vl = self._windSpeed[nLayer]
            vdir = self._windDirection[nLayer]
            tf, sp_perp = self._xp.meshgrid(temp_freqs, self._spat_freqs)

        f = self._xp.sqrt(sp_perp ** 2 + (tf / vl) ** 2)
        a1, a2, sl, thS, psd = self._initializeGeometry(nLayer, f)
        _, mj, _, mk, deltaj, deltak, b1, b2, c0, c1 = \
            self._initializeFunctionsForZernikeComputations(
                j, k, f, a1, a2)

        thWind = self._xp.deg2rad(vdir)
        th0 = self._xp.arccos(-tf / (f * vl))
        th1 = th0 + thWind
        th2 = -th0 + thWind

        c4 = self._xp.pi / 4 * (1 - deltaj) * ((-1) ** j - 1)
        c5 = self._xp.pi / 4 * (1 - deltak) * ((-1) ** k - 1)

        integFunc = c0 * c1 / (vl * self._xp.pi) * \
            psd / f ** 2 * b1 * b2 * \
            (self._xp.exp(-2 * 1j * self._xp.pi * f * sl * self._xp.cos(
                th1 - thS)) *
             self._xp.cos(mj * th1 + c4) *
             self._xp.cos(mk * th1 + c5) +
             self._xp.exp(-2 * 1j * self._xp.pi * f * sl * self._xp.cos(
                 th2 - thS)) *
             self._xp.cos(mj * th2 + c4) *
             self._xp.cos(mk * th2 + c5))

        if self._xp == cp:
            return self._xp.asnumpy(integFunc), self._xp.asnumpy(sp_perp)
        elif self._xp == np:
            return integFunc, sp_perp

    def integrandOfPhaseCPSD(self, nLayer, temp_freqs):
        if self._xp == cp:
            vl = self._xp.array(self._windSpeed[nLayer])
            vdir = self._xp.array(self._windDirection[nLayer])
            tf, sp_perp = self._xp.meshgrid(
                self._xp.array(temp_freqs), self._xp.array(self._spat_freqs))
        elif self._xp == np:
            vl = self._windSpeed[nLayer]
            vdir = self._windDirection[nLayer]
            tf, sp_perp = self._xp.meshgrid(temp_freqs, self._spat_freqs)

        f = self._xp.sqrt(sp_perp ** 2 + (tf / vl) ** 2)
        a1, a2, sl, thS, psd = \
            self._initializeGeometry(nLayer, f)
        self._initializeFunctionsForPhaseComputations(f, a1, a2)

        thWind = self._xp.deg2rad(vdir)
        th0 = self._xp.arccos(-tf / (f * vl))
        th1 = th0 + thWind
        th2 = -th0 + thWind

        intFunc = 1. / vl * psd * (
            self._b1Phase - self._b2Phase * self._b3Phase) * (
                self._xp.exp(-2 * 1j * self._xp.pi * f * sl * self._xp.cos(
                    th1 - thS)) +
            self._xp.exp(-2 * 1j * self._xp.pi * f * sl * self._xp.cos(
                th2 - thS)))

        if self._xp == cp:
            return self._xp.asnumpy(intFunc), self._xp.asnumpy(sp_perp)
        elif self._xp == np:
            return intFunc, sp_perp

    def integrandOfGeneralPhaseCPSD(self, nLayer, temp_freqs):
        if self._xp == cp:
            vl = self._xp.array(self._windSpeed[nLayer])
            vdir = self._xp.array(self._windDirection[nLayer])
            tf, sp_perp = self._xp.meshgrid(
                self._xp.array(temp_freqs), self._xp.array(self._spat_freqs))
#             tf = self._xp.array(temp_freqs)
        elif self._xp == np:
            vl = self._windSpeed[nLayer]
            vdir = self._windDirection[nLayer]
            tf, sp_perp = self._xp.meshgrid(temp_freqs, self._spat_freqs)
#             tf = temp_freqs

        f = self._xp.sqrt(sp_perp ** 2 + (tf / vl) ** 2)
#         def f(sf_perp):
#             tf_g, sf_g = np.meshgrid(tf, sf_perp)
#             return self._xp.sqrt(sf_g ** 2 + (tf_g / vl) ** 2)

        a1, a2, sl, thS, psd = \
            self._initializeGeometry(nLayer, f)
        self._initializeFunctionsForPhaseComputations(f, a1, a2)

        thWind = self._xp.deg2rad(vdir)
        th0 = self._xp.arccos(-tf / (f * vl))

#         def th0(f):
#             return self._xp.arccos(-tf / (f * vl))

#         def cos1(f):
#             return f**0 * self._xp.cos(
#                 2 * np.pi * f * sl * self._xp.cos(thWind - thS) *
#                 self._xp.cos(th0(f)))

#         def cos2(f):
#             return f**0 * self._xp.cos(
#                 2 * np.pi * f * sl * self._xp.sin(thWind - thS) *
#                 self._xp.sin(th0(f)))

        intFunc = 4. / vl * psd * (
            self._b1Phase - self._b2Phase * self._b3Phase
        ) * self._xp.cos(
            2 * np.pi * f * sl * self._xp.cos(thWind - thS) * self._xp.cos(
                th0)) * self._xp.cos(
            2 * np.pi * f * sl * self._xp.sin(thWind - thS) *
            self._xp.sin(th0))

        if self._xp == cp:
            return self._xp.asnumpy(intFunc), self._xp.asnumpy(sp_perp)
        elif self._xp == np:
            return intFunc, sp_perp
#         return lambda sf_perp: 4. / vl * self._VonKarmanPSDOneLayer(
#             nLayer, f(sf_perp)) * (
#             self._b1Phase(f(sf_perp)) - self._b2Phase(f(sf_perp)) *
#             self._b3Phase(f(sf_perp))) * \
#             cos1(f(sf_perp)) * cos2(f(sf_perp))

    def getZernikeCovariance(self, j, k):
        """
        Return the covariance between two Zernike coefficients with index j
        and k describing the phase seen, respectively, on aperture1 from
        source1 and on aperture2 from source2.

        Parameters
        ----------
        j: int or list
            Index of Zernike coefficients related to source1 on aperture1.

        k: int or list
            Index of Zernike coefficients related to source2 on aperture2.

        Returns
        -------
        zernikeCovariance: `~astropy.units.quantity.Quantity`
            Zernike covariance or covariance matrix (matrix of shape nxm if n
            and m are, respectively, the dimension of j and k) in [rad**2].
        """

        if (np.isscalar(j) and np.isscalar(k)):
            zernikeCovariance = self._computeZernikeCovarianceAllLayers(j, k)

        elif (np.isscalar(j) and np.isscalar(k) is False):
            zernikeCovariance = np.array([
                self._computeZernikeCovarianceAllLayers(j, k_mode)
                for k_mode in k])

        elif (np.isscalar(j) is False and np.isscalar(k)):
            zernikeCovariance = np.array([
                self._computeZernikeCovarianceAllLayers(j_mode, k)
                for j_mode in j])

        else:
            zernikeCovariance = np.array([
                [self._computeZernikeCovarianceAllLayers(j_mode, k_mode)
                    for k_mode in k]
                for j_mode in j])

        return zernikeCovariance * u.rad ** 2

    def getPhaseCovariance(self):
        """
        Return the covariance between the phase seen from source1 on
        aperture1 and the phase seen from source2 on aperture2.

        Returns
        -------
        phaseCovariance: `~astropy.units.quantity.Quantity`
            Covariance between phase1 and phase2 in [rad**2].
        """
        phaseCovAllLayers = np.array([
            self._computePhaseCovarianceOneLayer(nLayer) for nLayer
            in range(self._numberOfLayers)])
        phaseCovariance = phaseCovAllLayers.sum()
        return phaseCovariance * u.rad ** 2

    def getZernikeCPSD(self, j, k, temp_freqs):
        """
        Return the Cross Power Spectral Density (CPSD) of the Zernike
        coefficients with index j and k describing the phase seen on aperture1
        and aperture2 observing, respectively, source1 and source2.
        The CPSD is a function of the temporal frequency.

        Parameters
        ----------
        j: int
            Index of the Zernike coefficient (related to source1 on aperture1).
        k: int
            Index of the Zernike coefficient (related to source2 on aperture2).
        temp_freqs: numpy.ndarray
            Temporal frequencies array in Hz.

        Returns
        -------
        zernikeCPSD: `~astropy.units.quantity.Quantity`
            Zernike CPSD or matrix of Zernike CPSDs in [rad**2/Hz].
        """

        self._temporalFreqs = temp_freqs

        if (np.isscalar(j) and np.isscalar(k)):
            zernikeCPSD = self._getZernikeCPSDAllLayers(
                j, k, temp_freqs)

        elif (np.isscalar(j) and np.isscalar(k) is False):
            zernikeCPSD = np.array([
                self._getZernikeCPSDAllLayers(j, k_mode, temp_freqs)
                for k_mode in k])

        elif (np.isscalar(j) is False and np.isscalar(k)):
            zernikeCPSD = np.array([
                self._getZernikeCPSDAllLayers(j_mode, k, temp_freqs)
                for j_mode in j])

        else:
            zernikeCPSD = np.array([
                [self._getZernikeCPSDAllLayers(j_mode, k_mode, temp_freqs)
                    for k_mode in k]
                for j_mode in j])

        return zernikeCPSD * u.rad ** 2 / u.Hz

    def getGeneralZernikeCPSD(self, j, k, temp_freqs):
        # TODO: Is this function necessary? We can get the same result
        # computing 2 * np.real(getZernikeCPSD).
        """
        Return the generalized expression of Zernike CPSD that we get
        from 'getZernikeCPSD' function.
        This expression is needed when we have to integrate the Zernike CPSD
        in the temporal frequency range from -infinity to +infinity. Instead of
        this computation, we can obtain the same result performing the integral
        of the generalized Zernike CPSD in the temporal frequency range from
        0 to infinity. This is what we need, for example, when we want to
        compare the Zernike covariance with the Zernike CPSD integrated in
        the temporal frequencies' domain.

        Parameters
        ----------
        j: int or list
            Index of Zernike coefficients related to source1 on aperture1.
        k: int or list
            Index of Zernike coefficients related to source2 on aperture2.
        temp_freqs: numpy.ndarray
            Temporal frequencies array.

        Returns
        -------
        cpsdTotal: `~astropy.units.quantity.Quantity`
            General Zernike CPSD in [rad**2/Hz].
        """

        self._temporalFreqs = temp_freqs

        if (np.isscalar(j) and np.isscalar(k)):
            zernikeCPSD = self._getGeneralZernikeCPSDAllLayers(
                j, k, temp_freqs)

        elif (np.isscalar(j) and np.isscalar(k) is False):
            zernikeCPSD = np.array([
                self._getGeneralZernikeCPSDAllLayers(j, k_mode, temp_freqs)
                for k_mode in k])

        elif (np.isscalar(j) is False and np.isscalar(k)):
            zernikeCPSD = np.array([
                self._getGeneralZernikeCPSDAllLayers(j_mode, k, temp_freqs)
                for j_mode in j])

        else:
            zernikeCPSD = np.array([
                [self._getGeneralZernikeCPSDAllLayers(j_mode, k_mode,
                                                      temp_freqs)
                    for k_mode in k]
                for j_mode in j])

        return zernikeCPSD * u.rad ** 2 / u.Hz

    def getPhaseCPSD(self, temp_freqs):
        """
        Return the Cross Power Spectral Density (CPSD) of the turbulent phase
        seen by aperture1 and aperture2 observing, respectively, source1 and
        source2.
        The CPSD is a function of temporal frequency.

        Parameters
        ----------
        temp_freqs: numpy.ndarray
            Temporal frequencies array.

        Returns
        -------
        phaseCPSD: `~astropy.units.quantity.Quantity`
            Phase CPSD in [rad**2/Hz].
        """

        self._temporalFreqs = temp_freqs

        phaseCPSD = self._getPhaseCPSDAllLayers(temp_freqs) * u.rad ** 2 / u.Hz
        return phaseCPSD

    def getGeneralPhaseCPSD(self, temp_freqs):
        """
        """

        self._temporalFreqs = temp_freqs

        generalPhaseCPSD = self._getGeneralPhaseCPSDAllLayers(
            temp_freqs) * u.rad ** 2 / u.Hz
        return generalPhaseCPSD

    def plotCPSD(self, cpsd, func_part, scale, legend):
        import matplotlib.pyplot as plt
#         if wavelength is None:
#             lam = self._cn2.DEFAULT_LAMBDA
#         else:
#             lam = wavelength
        lam = self._cn2.wavelength().value
        temp_freqs = self.temporal_frequencies()
        m_to_nm = 1e18
        if func_part == 'real':
            if scale == 'log':
                plt.loglog(
                    temp_freqs,
                    np.abs(np.real(cpsd)) *
                    (lam / (2 * np.pi)) ** 2 * m_to_nm,
                    '-', label=legend)
            elif scale == 'linear':
                plt.semilogx(
                    temp_freqs,
                    np.real(cpsd) *
                    (lam / (2 * np.pi)) ** 2 * m_to_nm,
                    '-', label=legend)
        elif func_part == 'imag':
            if scale == 'log':
                plt.loglog(
                    temp_freqs,
                    np.abs(np.imag(cpsd)) *
                    (lam / (2 * np.pi)) ** 2 * m_to_nm,
                    '-', label=legend)
            elif scale == 'linear':
                plt.semilogx(
                    temp_freqs,
                    np.imag(cpsd) *
                    (lam / (2 * np.pi)) ** 2 * m_to_nm,
                    '-', label=legend)
        plt.legend()
        plt.xlabel('Frequency [Hz]')
        plt.ylabel('CPSD [nm$^{2}$/Hz]')



from itertools import chain  # List flattening

class BaseIndexer():

    def __init__(self, single_kw=None, from_kw=None, to_kw=None):
        self.single_kw = ['element', 'elements']
        self.from_kw= ['from_element', 'first']
        self.to_kw = ['to_element', 'last']

        if single_kw:
            self.single_kw += [single_kw]
            self.single_kw = list(chain.from_iterable(self.single_kw))
        if from_kw:
            self.from_kw += [from_kw]
            self.from_kw = list(chain.from_iterable(self.from_kw))
        if to_kw:
            self.to_kw += [to_kw]
            self.to_kw = list(chain.from_iterable(self.to_kw))

    def process_args(self, *args, **kwargs):
        '''
        default: all elements
        element = single element
        elements = list of elements
        from_element = first element
        to_element = last element
        '''
        elements = None
        from_element = None
        to_element = None
        if len(args) > 0:
            elements = args
        else:
            for k, v in kwargs.items():
                if k in self.single_kw:
                    elements = v
                if k in self.from_kw:
                    from_element = v
                if k in self.to_kw:
                    to_element = v
        if elements is not None:
            return elements
        return slice(from_element, to_element)


import numpy as np
import struct



class FisbaMeasure(object):

    TYPE_SURFACE_DEVIATION= 1
    TYPE_WAVE_ABERRATION= 2
    TYPE_RAW_PHASE_DATA= 3
    TYPE_INTENSITY= 4


    def __init__(self, filename):
        self._filename= filename
        self._dataMap, self._dataType, self._comment= self._convert(
            self._filename)


    def _convert(self, filename):
        with open(filename, 'rb') as binary_file:
            data= bytearray(binary_file.read())
        dataId, nRow, nCol, maskValue, commentLen=struct.unpack(
            "<ciiii", data[0:17])
        dataType=ord(dataId)
        comment=struct.unpack(
            "<%ds" % commentLen,
            data[17:17 + commentLen])[0]
        dataArray=np.array(struct.unpack(
            "<" + "i"* nRow * nCol,
            data[17 + commentLen:17 + commentLen + nRow * nCol * 4])
        ).reshape((nRow, nCol))
        mask= dataArray == maskValue
        return np.ma.array(dataArray, mask=mask), dataType, comment


    def dataType(self):
        return self._dataType

    def map(self):
        return self._dataMap

    def comment(self):
        return self._comment



import numpy as np

"""
From J. Negro 'Subaperture optical system testing' 10.1364/ao.23.001921

NB: Negro84 doesn't use the same ordering of Zernike of Noll76. This
class reorder the coefficients according to Noll76

BUG: something wrong in computed values z2,z3 when z7 is not null in the full
pupil
"""


class ZernikeProjectionOnSubaperture(object):
    
    N_MODES = 10

    def __init__(self,
                 pupilRadiusInMeter,
                 subapsRadiusInMeter,
                 subapOffAxisRadiusInMeter,
                 subapOffAxisAzimuthInDegrees):
        self._R = pupilRadiusInMeter
        self._r = subapsRadiusInMeter
        self._h = subapOffAxisRadiusInMeter
        self._a = subapOffAxisAzimuthInDegrees
        self._S = self._computeMatrix()

    def compute_zernike_decomposition_on_subap(self, zernikeCoeffFrom2To11):
        return np.dot(zernikeCoeffFrom2To11, self._S)

    def get_projection_matrix(self):
        return self._S
    

    def _computeMatrix(self):
        roR = self._r / self._R
        hoR = self._h / self._R
        ca = np.cos(self._a * np.pi / 180)
        sa = np.sin(self._a * np.pi / 180)
        c2a = np.cos(2 * self._a * np.pi / 180)
        s2a = np.sin(2 * self._a * np.pi / 180)

        z2 = np.array([roR, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        z3 = np.array([0, roR, 0, 0, 0, 0, 0, 0, 0, 0])
        z4 = np.array([2 * np.sqrt(3) * hoR * roR * ca,
                       2 * np.sqrt(3) * hoR * roR * sa,
                       roR ** 2, 0, 0, 0, 0, 0, 0, 0])
        z6 = np.array([np.sqrt(6) * hoR * roR * ca,
                       -np.sqrt(6) * hoR * roR * sa,
                       0, 0, roR ** 2, 0, 0, 0, 0, 0])
        z5 = np.array([np.sqrt(6) * hoR * roR * sa,
                       np.sqrt(6) * hoR * roR * ca,
                       0, roR ** 2, 0, 0, 0, 0, 0, 0])
        z8 = np.array([3 * np.sqrt(2) * roR * (hoR ** 2 * (1 + 2 * ca ** 2) + 2. / 3 * roR ** 2 - 2. / 3),
                       3 * np.sqrt(2) * hoR ** 2 * roR * s2a,
                       2 * np.sqrt(6) * hoR * roR ** 2 * ca,
                       2 * np.sqrt(3) * hoR * roR ** 2 * sa,
                       2 * np.sqrt(3) * hoR * roR ** 2 * ca,
                       0, roR ** 3, 0, 0, 0])
        z7 = np.array([3 * np.sqrt(2) * hoR ** 2 * roR * s2a,
                       3 * np.sqrt(2) * roR * (hoR ** 2 *
                                               (1 + 2 * sa ** 2) + 2. / 3 * roR ** 2 - 2. / 3),
                       2 * np.sqrt(6) * hoR * roR ** 2 * sa,
                       2 * np.sqrt(3) * hoR * roR ** 2 * ca,
                       -2 * np.sqrt(3) * hoR * roR ** 2 * sa,
                       roR ** 3, 0, 0, 0, 0])
        z10 = np.array([3 * np.sqrt(2) * hoR ** 2 * roR * c2a,
                        -3 * np.sqrt(2) * hoR ** 2 * roR * s2a,
                        0,
                        -2 * np.sqrt(3) * hoR * roR ** 2 * sa,
                        2 * np.sqrt(3) * hoR * roR ** 2 * ca,
                        0, 0, 0, roR ** 3, 0])
        z9 = np.array([3 * np.sqrt(2) * hoR ** 2 * roR * s2a,
                       3 * np.sqrt(2) * hoR ** 2 * roR * c2a,
                       0,
                       2 * np.sqrt(3) * hoR * roR ** 2 * ca,
                       2 * np.sqrt(3) * hoR * roR ** 2 * sa,
                       0, 0, roR ** 3, 0, 0])
        z11 = np.array([12 * np.sqrt(5) * hoR * roR * (hoR ** 2 + 2. / 3 * roR ** 2 - 0.5) * ca,
                        12 * np.sqrt(5) * hoR * roR * (hoR ** 2 +
                                                       2. / 3 * roR ** 2 - 0.5) * sa,
                        (4 * hoR ** 2 + roR ** 2 - 1) * np.sqrt(15) * roR ** 2,
                        2 * np.sqrt(30) * hoR ** 2 * roR ** 2 * s2a,
                        2 * np.sqrt(30) * hoR ** 2 * roR ** 2 * c2a,
                        2 * np.sqrt(10) * hoR * roR ** 3 * sa,
                        2 * np.sqrt(10) * hoR * roR ** 3 * ca,
                        0, 0, roR ** 4])
        return np.vstack((z2, z3, z4, z5, z6, z7, z8, z9, z10, z11))



import numpy as np
from arte.types.domainxy import DomainXY
from arte.utils.help import add_help
from arte.utils.radial_profile import computeRadialProfile
from scipy import interpolate
from scipy.interpolate import LinearNDInterpolator, RegularGridInterpolator

@add_help
class ScalarBidimensionalFunction(object):
    '''
    Represents a scalar function in an XY plane

    The function is initialized with a 2d value and one of the following:
        - a domain over which it is defined
        - two X and Y maps with the domain coordinates
        - (nothing passed) a default domain with the same shape as the value,
          with centered origin and unitary step

    Parameters
    ----------
    values_array: numpy.ndarray
       two-dimensional array with the function value
    xmap: numpy.ndarray, optional
       two-dimensional array with the X coordinate at which each function
       value is sampled.
    ymap: numpy.ndarray, optional
       two-dimensional array with the Y coordinate at which each function
       value is sampled.
    domain: DomainXY instance, optional
       the domain over which the function is sampled.

    Raises
    ------
    ValueError
       if the input parameters do not satisfy the requirements, for example:
           - xmap is passed, but not ymap
           - both xmap/ymap and a domain has been passed
           - values_array is not a 2d array
           - shapes of values and xmap/ymap or domain don't match
    '''

    def __init__(self, values_array, xmap=None, ymap=None, domain=None):

        if len(values_array.shape) != 2:
            raise ValueError('values_array must be 2d')

        if (xmap is not None) or (ymap is not None):
            if (ymap is None) or (xmap is None):
                raise ValueError('xmap and ymap must be specified together')
            if domain is not None:
                raise ValueError('both domain and x/y maps specified')

            domain = DomainXY.from_xy_maps(xmap, ymap)
        else:
            if not domain:
                domain = DomainXY.from_shape(values_array.shape, 1)

        self._values = values_array
        self._domain = domain
        self._check_passed_arrays()

        self.xcoord = self._domain.xcoord
        self.ycoord = self._domain.ycoord
        self.xmap = self._domain.xmap
        self.ymap = self._domain.ymap
        self.extent = self._domain.extent
        self.origin = self._domain.origin

    def _check_passed_arrays(self):
        self._check_shapes()

    def _check_shapes(self):
        if not self.values.shape == self._domain.shape:
            raise ValueError(
                "values shape (%s) don't match domain shape (%s)" %
                (str(self.values.shape), str(self._domain.shape)))

    @property
    def values(self):
        '''2d values array'''
        return self._values

    @property
    def domain(self):
        '''DomainXY instance'''
        return self._domain

    @property
    def shape(self):
        '''(y,x) function shape'''
        return self._values.shape

    def interpolate_in_xy(self, x, y, span=3):
        '''
        Interpolate the function at points x,y

        Returns an array of z[i] corresponding to the values interpolated
        at points x[i], y[i]

        The interpolation algorithm is linear and uses a box of 3x3 points
        centered on the closest point to the interpolation coordinate


        Parameters
        ----------
        x: numpy.ndarray or tuple with size N
            1D array of the x coordinates of the sampling points

        y: numpy.ndarray or tuple with size N
            1D array of the y coordinates of the sampling points

        Returns
        -------
        z: numpy.ndarray
            array of shape (N,) of the sampled values. z[i] = interp(f, x[i], y[i]) 

        '''
        return self._my_interp(x, y, span=span)

    def get_radial_profile(self):
        '''Get the radial profile around the domain origin.

        Assumes that the domain sampling is the same in the X and Y
        directions
        '''
        radial_profile, radial_distance_in_px = computeRadialProfile(
            self._values, self._domain.origin[0], self._domain.origin[1])
        return radial_profile, radial_distance_in_px * self._domain.step[0]

    def plot_radial_profile(self):
        import matplotlib.pyplot as plt
        y, x = self.get_radial_profile()
        plt.plot(x, y)
        plt.show()

    def _my_interp(self, x, y, span=3):
        xs, ys = map(np.array, (np.atleast_1d(x), np.atleast_1d(y)))
        z = np.zeros(xs.shape)
        for i, (x, y) in enumerate(zip(xs, ys)):
            # get the indices of the nearest x,y
            box = self._domain.get_boundingbox_slice(x, y, span=span)

            # make slices of X,Y,Z that are only a few items wide
            nX = self._domain.xmap[box]
            nY = self._domain.ymap[box]
            nZ = self.values[box]

            if np.iscomplexobj(nZ):
                z[i] = self._interp_complex(nX, nY, nZ, x, y)
            else:
                z[i] = self._interp_real(nX, nY, nZ, x, y)
        return z

    def _interp_real_obsolete(self, nX, nY, nZ, x, y):
        # beware of https://github.com/scipy/scipy/issues/10268 !!!
        # use 1D (nX[0] and nY[:,0]) instead of 2D nX and nY
        intp = interpolate.interp2d(nX[0], nY[:, 0], nZ, kind='linear')
        return intp(x, y)[0]

    def _interp_complex_obsolete(self, nX, nY, nZ, x, y):
        intpr = self._interp_real_obsolete(
            nX[0], nY[:, 0], nZ.real, kind='linear')
        intpi = self._interp_real_obsolete(
            nX[0], nY[:, 0], nZ.imag, kind='linear')
        return intpr(x, y)[0] + 1j * intpi(x, y)[0]

    def _interp_real_linear_nd(self, nX, nY, nZ, x, y):
        xy = np.c_[nX[0], nY[:, 0]]
        r = LinearNDInterpolator(xy, nZ)
        return r(x, y)

    def _interp_real_regular_grid_interpolator(self, nX, nY, nZ, x, y):
        xy = (nY[:, 0], nX[0])
        r = RegularGridInterpolator(
            xy, nZ, bounds_error=False, fill_value=None)
        return r((y, x))

    def _interp_complex_regular_grid_interpolator(self, nX, nY, nZ, x, y):
        zr = self._interp_real_regular_grid_interpolator(nX, nY, nZ.real, x, y)
        zi = self._interp_real_regular_grid_interpolator(nX, nY, nZ.imag, x, y)
        return zr + 1j * zi

    def _interp_real(self, nX, nY, nZ, x, y):
        return self._interp_real_regular_grid_interpolator(nX, nY, nZ, x, y)

    def _interp_complex(self, nX, nY, nZ, x, y):
        return self._interp_complex_regular_grid_interpolator(nX, nY, nZ, x, y)

    def get_roi(self, xmin, xmax, ymin, ymax):

        box = self._domain.get_crop_slice(xmin, xmax, ymin, ymax)
        cropped_values = self.values[box]
        cropped_domain = self._domain[box]
        return ScalarBidimensionalFunction(cropped_values,
                                           domain=cropped_domain)



import abc

class AbstractFileNameWalker(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def snapshot_dir(self, tag):
        '''
        Abstract method that must be reimplemented in the derived class.

        Parameters
        ----------
        tag: str
            Unique tag that identifies an analyzer data directory

        Returns
        -------
        fullpath: str
            Full path including the tag directory
        '''
        pass

    def find_tag_between_dates(self, tag_start, tag_stop):
        '''Return the list of tags between the two extremes'''
        return []



import numpy as np


class Constants(object):
    '''Class holding global constants'''

    RAD2ARCSEC = 180 * 3600 / np.pi
    ARCSEC2RAD = np.pi / (180 * 3600)
    DEG2RAD = np.pi / 180


RAD2ARCSEC = 180 * 3600 / np.pi
ARCSEC2RAD = np.pi / (180 * 3600)
DEG2RAD = np.pi / 180



#!/usr/bin/env python3
#
#########################################################
#
# who       when        what
# --------  ----------  ---------------------------------
# apuglisi  2020-07-24  Created
#
#########################################################
'''
Utilites for CPU/GPU agnostic code using CUPY.

All functions in this module are safe to call even when cupy is not installed,
because they will use cupy only when an user-supplied value is a cupy
array (and thus cupy has already been imported).
'''

from functools import wraps


def is_numpy_or_cupy_array(arr):
    '''Returns True if the argument is either a numpy or a cupy array

    The check is performed looking at object attributes, therefore
    can be called without importing the cupy module.
    '''
    return hasattr(arr, '__array_function__') and \
           hasattr(arr, '__array_ufunc__')
    

def is_cupy_array(arr):
    '''Returns True if the argument is a cupy array.

    The check is performed looking at object attributes, therefore
    can be called without importing the cupy module.
    '''
    return hasattr(arr, '__array_function__') and \
           hasattr(arr, '__array_ufunc__') and \
           hasattr(arr, 'device')


def is_numpy_array(arr):
    '''Returns True if the argument is a numpy array.

    The check is performed looking at object attributes, therefore
    can be called without importing the cupy module.
    '''
    return hasattr(arr, '__array_function__') and \
           hasattr(arr, '__array_ufunc__') and \
           not hasattr(arr, 'device')


def get_array_module(arr):
    '''
    A reimplementation of cupy.get_array_module() that works
    also when cupy is not installed.
    '''
    if is_cupy_array(arr):
        import cupy
        return cupy
    else:
        import numpy
        return numpy


def _format_owner(owner):

    if owner is None:
        return 'None'

    owner = owner.__class__.__name__
    if hasattr(owner, 'name'):
       owner += '(%s)' % owner.name
    return owner


def from_GPU(value, owner=None, quiet=False):
    '''
    If *value* is a cupy array, transfer it to the host
    and return the corresponding numpy array.

    All other values are left alone. It is safe to call
    this function with any kind of object, and also when
    cupy is not installed. In the latter case, of course
    no transfer will be done.
    '''
    if is_cupy_array(value):

        if not quiet:
            import cupy
            infostr = 'from_GPU: transferring %s %s' % (value.shape, value.dtype)
            infostr += ' from device %d' % cupy.cuda.runtime.getDevice()
            infostr += ' (owner %s)' % _format_owner(owner)
            print(infostr)
        return value.get()
    else:
        return value


def to_GPU(array, owner=None, quiet=False):
    '''
    Transfer *array* to the GPU and return the corresponding cupy array.
    '''
    if array is None:
        return array

    if not is_cupy_array(array):
        import cupy

        if not quiet:
            if is_numpy_array(array):
                infostr = 'to_GPU: transferring %s %s' % (array.shape, array.dtype)
            else:
                infostr = 'to_GPU: transferring object %s' % str(array)
            infostr += ' to device %d' % cupy.cuda.runtime.getDevice()
            infostr += ' (owner %s)' % _format_owner(owner)
            print(infostr)

        array = cupy.array(array)
    return array


def _make_sure_first_arg_is_little_endian(f):
    '''
    Decorator to enforce little endianess with cupy.

    This is a workaround for the cupy issue with FITS files:
    https://github.com/cupy/cupy/issues/3652

    It will analyze the first argument and, if it is
    a big-endian array (like the ones returned by astropy.io.fits.getdata),
    convert it to a little-endian one before passing it to the function.

    Can be used to "patch" the cupy routines as follows::

        cupy.array = _make_sure_first_arg_is_little_endian(cupy.array)
        cupy.asarray = _make_sure_first_arg_is_little_endian(cupy.asarray)
    '''
    @wraps(f)
    def wrapper(data, *args, **kwargs):
        import sys
        import numpy

        if isinstance(data, numpy.ndarray):
            # Use names compatible with sys.byteorder
            endianess_map = {
                '>': 'big',
                '<': 'little',
                '=': sys.byteorder,
                '|': 'not relevant'
            }
            endianess = endianess_map[data.dtype.byteorder]
            if endianess == 'big':
               print('Array %s %s must be converted from big to little endian' %
                     (str(data.dtype), str(data.shape)))
               data = data.astype(data.dtype.newbyteorder('<'))

        return f(data, *args, **kwargs)

    return wrapper


def cupy_patch(cupy):
    '''
    Patch the passed cupy module with our workarounds for FITS files.
    '''
    cupy.array = _make_sure_first_arg_is_little_endian(cupy.array)
    cupy.asarray = _make_sure_first_arg_is_little_endian(cupy.asarray)    


import threading
import queue

class WorkerThread:
    '''
    Generic worker thread, using queues to communicate.

    After being initialized with *func*, calling *trigger()*
    will queue values to be processed by *func*, while
    *result()* will fetch the next result from *func*.

    *result()* may block until the next result is available.

    The thread will exit once all references to it go out
    of scope. Call *terminate()* to exit the thread explicitly.
    '''

    sentinel = object()

    def __init__(self, func):
        self._qin = queue.Queue()
        self._qout = queue.Queue()
        self._func = func
        self.thread = threading.Thread(target=self.worker_loop, daemon=True)
        self.thread.start()

    def __del__(self):
        self.terminate()

    def worker_loop(self):
        while True:
            value = self._qin.get()
            if value is self.sentinel:
                return
            result = self._func(value)
            self._qout.put(result)

    def trigger(self, value):
        self._qin.put(value)

    def result(self):
        return self._qout.get()

    def terminate(self):
        self._qin.put(self.sentinel)


# ___oOo___





def _paste_slices(tup):
    pos, w, max_w = tup
    wall_min = max(pos, 0)
    wall_max = min(pos+w, max_w)
    block_min = -min(pos, 0)
    block_max = max_w-max(pos+w, max_w)
    block_max = block_max if block_max != 0 else None
    return slice(wall_min, wall_max), slice(block_min, block_max)


def paste(wall, block, loc):
    '''
    Paste `block` array into `wall` in location `loc`.
    
    `loc` must be a tuple (even if block.ndim=1)
    Going outside of the edges of `wall` is properly managed.
    `wall` is modified in place.
    
    Parameters
    ----------
    wall: `~numpy.array` of ndim=N
        Array in which `block` is pasted.
    block: `~numpy.array` of ndim=N
        Array to be pasted into `wall`
    loc: tuple, shape (N,)
        location where to paste `block`.

    Returns
    -------
    None

    Example
    -------
    Insert array [1,2,3,4] into an empty array of size 10 at location 8

    >>> b = np.zeros([10])
    >>> a = np.arange(1, 5)
    >>> b
    array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.])
    >>> a
    array([1, 2, 3, 4])
    >>> paste(b, a, (8,))
    >>> b
    array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,  2.])

    References
    ----------
    https://stackoverflow.com/questions/7115437/how-to-embed-a-small-numpy-array-into-a-predefined-block-of-a-large-numpy-arra
    '''

    loc_zip = zip(loc, block.shape, wall.shape)
    wall_slices, block_slices = zip(*map(_paste_slices, loc_zip))
    wall[wall_slices] = block[block_slices]



# -*- coding: utf-8 -*-

import numpy as np
import astropy.units as u

from arte.utils.help import add_help
from arte.math.make_xy import make_xy
from arte.utils.unit_checker import assert_unit_is_equivalent, \
                                    separate_value_and_unit


def _accept_one_or_two_elements(x, name=''):

    errmsg = '%s must be either a scalar or a 2-elements sequence' % name

    # Special case for astropy Quantities, who have __len__
    # but throw exceptions if they are scalars.
    if isinstance(x, u.Quantity) and x.isscalar:
        x = (x, x)

    # Make sure it's a sequence, but not a string, which also has a __len__
    if not hasattr(x, '__len__'):
        x = (x, x)
    elif isinstance(x, str):
        raise TypeError(errmsg)

    if len(x) != 2:
        raise TypeError(errmsg)

    return x


@add_help
class DomainXY():
    '''Holds information about a 2d domain

    For users of the old IDL make_xy, this is the same in class form.
    Several initializers are provided to build a domain from different
    data.

    Domains support astropy units and try to play nicely with them. For
    example, the `shift()` method will shift a domain using the correct unit
    if both the domain and the shift parameter has one. If not, everything
    is converted to unitless, shifted, and then the unit is applied again.
    Shapes and indexes in Python are strictly integer data type and cannot
    use a unit.

    Domains can be compared with == and !=, but no ordering is defined
    between them.

    In addition to the access methods, domain data can be
    Since the domain sampling is regular, the class internally only stores two
    linear vectors for the X and Y sampling. Everything else is dynamically
    calculated on the fly from these vectors.
    '''

    def __init__(self, xcoord_vector, ycoord_vector):
        self._xcoord = xcoord_vector.copy()
        self._ycoord = ycoord_vector.copy()
        assert len(self._xcoord.shape) == 1
        assert len(self._ycoord.shape) == 1

    @classmethod
    def from_xy_vectors(cls, x_vector, y_vector):
        '''Build a domain as a cartesian product of two coordinate vectors'''
        return cls(x_vector, y_vector)

    @classmethod
    def from_extent(cls, xmin, xmax, ymin, ymax, npoints):
        '''Build a domain from a bounding box'''

        npoints = _accept_one_or_two_elements(npoints, 'npoints')

        x = np.linspace(xmin, xmax, npoints[0])
        y = np.linspace(ymin, ymax, npoints[1])
        return cls(x, y)

    @classmethod
    def from_shape(cls, shape, pixel_size=1):
        '''Build a domain from a shape and a pixel size'''

        pixel_size = _accept_one_or_two_elements(pixel_size, 'pixel_size')

        tot_size = (shape[0] * pixel_size[0], shape[1] * pixel_size[1])

        # a shape is (rows, cols), so equivalent to (y,x)
        y = np.linspace(-(tot_size[0] - pixel_size[0]) / 2,
                        (tot_size[0] - pixel_size[0]) / 2,
                        shape[0])

        x = np.linspace(-(tot_size[1] - pixel_size[1]) / 2,
                        (tot_size[1] - pixel_size[1]) / 2,
                        shape[1])

        return cls.from_xy_vectors(x, y)

    @classmethod
    def from_makexy(cls, *args, **kwargs):
        '''Same arguments as make_xy'''
        x, y = make_xy(*args, **kwargs)
        return cls.from_xy_maps(x, y)

    @classmethod
    def from_linspace(cls, *args, **kwargs):
        '''Cartesian product of two identical np.linspace(). Same arguments'''
        v = np.linspace(*args, **kwargs)
        return cls.from_xy_vectors(v, v)

    @classmethod
    def from_xy_maps(cls, xmap, ymap):
        '''Build a domain from two 2d maps (like the ones from make_xy)'''
        xcoord_vector = xmap[0, :]
        ycoord_vector = ymap[:, 0]
        return cls(xcoord_vector, ycoord_vector)

    @property
    def shape(self):
        '''(y,x) domain shape'''
        return (self.ycoord.shape[0], self.xcoord.shape[0])

    @property
    def xcoord(self):
        '''X coordinates vector'''
        return self._xcoord

    @property
    def ycoord(self):
        '''Y coordinates vector'''
        return self._ycoord

    @property
    def xmap(self):
        '''X coordinates 2d map'''
        return np.tile(self.xcoord, (self.ycoord.size, 1))

    @property
    def ymap(self):
        '''Y coordinates 2d map'''
        return np.tile(self.ycoord, (self.xcoord.size, 1)).T

    @property
    def radial_map(self):
        '''(r, angle [radians]) 2d map'''
        return (np.hypot(self.ymap, self.xmap),
                np.arctan(self.ymap, self.xmap) * u.rad)

    @property
    def step(self):
        '''(dx,dy) = pixel size'''
        return (self._compute_step_of_uniformly_spaced_vector(self.xcoord),
                self._compute_step_of_uniformly_spaced_vector(self.ycoord))

    def __eq__(self, other):

        return np.allclose(self._xcoord, other._xcoord) and \
               np.allclose(self._ycoord, other._ycoord)

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def origin_in_px(self):
        '''(x,y) = indexes of 0,0 coordinate, interpolated'''
        return (self._interpolate_for_value(self.xcoord, 0.0),
                self._interpolate_for_value(self.ycoord, 0.0))

    @property
    def origin(self):
        '''(x,y) = indexes of 0,0 coordinate, interpolated'''
        return self.origin_in_px

    @property
    def extent(self):
        '''[xmin, xmax, ymin, ymax] = minimum and maximum coordinates'''
        return [self.xcoord.min(), self.xcoord.max(),
                self.ycoord.min(), self.ycoord.max()]

    @property
    def unit(self):
        '''(x,y) = units used on X and Y, or 1 otherwise'''
        _, xunit = separate_value_and_unit(self._xcoord)
        _, yunit = separate_value_and_unit(self._ycoord)
        return (xunit, yunit)

    def shift(self, dx, dy):
        '''Shift the domain in place'''

        assert_unit_is_equivalent(dx, ref=self._xcoord)
        assert_unit_is_equivalent(dy, ref=self._ycoord)

        self._xcoord += dx
        self._ycoord += dy

    def shifted(self, dx, dy):
        '''Returns a new shifted domain'''

        new = self[:]
        new.shift(dx, dy)
        return new

    def contains(self, x, y):
        '''Returns True if the coordinates are inside the domain'''

        assert_unit_is_equivalent(x, ref=self._xcoord)
        assert_unit_is_equivalent(y, ref=self._ycoord)

        return (x >= self._xcoord.min()) and (x <= self._xcoord.max()) and \
               (y >= self._ycoord.min()) and (y <= self._ycoord.max())

    def restrict(self, x, y):
        '''(x,y) = new coordinates restricted to be inside this domain'''

        assert_unit_is_equivalent(x, ref=self._xcoord)
        assert_unit_is_equivalent(y, ref=self._ycoord)

        x = max(x, self._xcoord.min())
        y = max(y, self._ycoord.min())
        x = min(x, self._xcoord.max())
        y = min(y, self._ycoord.max())

        return x, y

    def get_boundingbox_slice(self, x, y, span=1):
        '''Slice that includes (x,y) with "span" pixels around'''

        x, y = self.restrict(x, y)

        idx_x = int(round(self._interpolate_for_value(self._xcoord, x)))
        idx_y = int(round(self._interpolate_for_value(self._ycoord, y)))

        xlo, xhi, ylo, yhi = idx_x - span, idx_x + span, \
                             idx_y - span, idx_y + span

        xlo = max(xlo, 0)
        ylo = max(ylo, 0)
        xhi = min(xhi, self._xcoord.size)
        yhi = min(yhi, self._ycoord.size)

        # Slices are (row,col), that is (y,x)
        return np.s_[ylo:yhi, xlo:xhi]

    def _indices(self, xmin, xmax, ymin, ymax, boundary_check=True):
        '''Returns the outer indices that correspond to a bounding box'''

        # Indexes must be unitless
        assert_unit_is_equivalent(xmin, ref=self._xcoord)
        assert_unit_is_equivalent(xmax, ref=self._xcoord)
        assert_unit_is_equivalent(ymin, ref=self._ycoord)
        assert_unit_is_equivalent(ymax, ref=self._ycoord)

        xlo = np.argmin(np.abs(self._xcoord - xmin))
        xhi = np.argmin(np.abs(self._xcoord - xmax)) + 2
        ylo = np.argmin(np.abs(self._ycoord - ymin))
        yhi = np.argmin(np.abs(self._ycoord - ymax)) + 2

        if boundary_check:
            xlo = max(xlo, 0)
            ylo = max(ylo, 0)
            xhi = min(xhi, self._xcoord.size)
            yhi = min(yhi, self._ycoord.size)

        return xlo, xhi, ylo, yhi

    def get_crop_slice(self, xmin, xmax, ymin, ymax):
        '''Slice to crop the domain'''
        xlo, xhi, ylo, yhi = self._indices(xmin, xmax, ymin, ymax)

        # Slices are (row,col), that is (y,x)
        return np.s_[ylo:yhi, xlo:xhi]

    def __getitem__(self, idx):
        '''Returns another DomainXY slicing this one'''
        # A single xx slice means (xx, :)
        if isinstance(idx, slice):
            idx = (idx, np.s_[:])

        xmap = self.xmap[idx]
        ymap = self.ymap[idx]
        return DomainXY.from_xy_maps(xmap, ymap)

    def cropped(self, xmin, xmax, ymin, ymax):
        '''Returns a new cropped DomainXY object'''
        xlo, xhi, ylo, yhi = self._indices(xmin, xmax, ymin, ymax)
        xc = self.xcoord[xlo:xhi]
        yc = self.ycoord[ylo:yhi]
        return DomainXY.from_xy_vectors(xc, yc)

    def boundingbox(self, x, y, span=1):
        '''Returns a new domain around the specified point'''
        slice_ = self.get_boundingbox_slice(x, y, span)
        return self[slice_]

    @staticmethod
    def _interpolate_for_value(arr, value):
        '''Returns the interpolated index (unitless) in `arr` for `value`'''

        return np.interp(value, arr, np.arange(arr.shape[0]),
                         left=np.nan, right=np.nan)

    @staticmethod
    def _compute_step_of_uniformly_spaced_vector(vector):
        delta = vector[1:] - vector[:-1]

        # np.isclose() does not like quantities
        d0, p2v = delta[0], np.ptp(delta)
        if isinstance(p2v, u.Quantity):
            d0, p2v = d0.value, p2v.value
        assert np.isclose(0, p2v, atol=1e-6 * d0)
        return delta[0]



'''
Created on 3 mag 2022

@author: giuliacarla
'''

import numpy as np
from scipy.special import gamma
from mpmath import hyper, fp
from arte.utils.zernike_generator import ZernikeGenerator


class VonKarmanSpatialCovariance():
    '''
    This function computes covariance of Zernike components of wavefront
    perturbations due to Von Karman turbulence. It returns covariance 
    <a(j1)*a(j2)> where a(j1) and a(j2) are j1-th and j2-th Zernike components
    of wavefront perturbations due to Von Karman turbulence. Covariance is
    in (wavelength)**2*(D/r0)**(5/3) units, where D is the pupil diameter and
    r0 the Fried parameter. To have covariance of phase perurbations, use:
        
        4*pi**2*VON_COVAR(j1, j2) = <2*pi*a(j1) * 2*pi*a(j2)>.
    
        It is in (D/r0)**(5/3) units (rad**2)
    
    Parameters
    ----------
    j1: int
        Index of Zernike polynomial. j1 >= 1
    j2: int
        Index of Zernike polynomial. j2 >= 1
    L0: float
        Outerscale value in pupil diameter units.
        
    References
    ---------- 
    Winker D. M., 1991, JOSA, 8, 1569.
    Takato N. and Yamaguchi I., 1995, JOSA, 12, 958
    
    Examples
    -------- 
    Computes the average spatial variance of phase perturbation on a 
    circular pupil of diameter 1.5m due to the first 15 Zernike aberrations.
    Suppose r0 = 15cm, L0=40m.
    
    scale = (150./15)**(5./3)
    L0norm= (40./1.5)
    var = 0.
    for i in range(2,16):
        vk = VonKarmanSpatialCovariance(i, i, L0norm)
        var = var + vk.get_covariance()
        var = 4 * np.pi**2 * var * scale
    Strhel_Ratio = np.exp(-var) 
    '''
    
    def __init__(self, j1, j2, L0):
        self._j1 = j1
        self._j2 = j2
        self._L0 = L0
        self._n1, self._m1 = ZernikeGenerator.degree(j1)
        self._n2, self._m2 = ZernikeGenerator.degree(j2)
        
    def _compute_covariance(self):
        n1pn2 = (self._n1 + self._n2) / 2
        n1mn2 = (self._n1 - self._n2)/2
        x0 = np.pi / self._L0    
        
        
        result = (
            24./5.*gamma(6./5))**(5./6)*(gamma(11./6))**2/np.pi**(5./2)
        result = result*(-1)**(
            (self._n1 + self._n2 - 2 * self._m1) / 2
            ) * np.sqrt((self._n1 + 1) * (self._n2 + 1))
        result = result / np.sin(np.pi * (n1pn2 + 1./6))
        
        temp1  = np.sqrt(np.pi) / 2**(
            self._n1 + self._n2 + 3) * gamma(n1pn2 + 1) / gamma(n1pn2 + 1./6)
        temp1  = temp1 / gamma(11./6) / gamma(self._n1 + 2.) / gamma(
            self._n2 + 2.) * x0**(self._n1 + self._n2 - 5./3.)

        alpha  = [n1pn2 + 3./2, n1pn2 + 2, n1pn2 + 1]
        beta   = [n1pn2 + 1./6, self._n1 + 2, self._n2 + 2, 
                  self._n1 + self._n2 + 3]
        
        temp1  = temp1 * hyper(alpha, beta, x0**2)

        temp2  = - gamma(7./3) * gamma(17./6) / 2
        temp2  = temp2 / gamma(11./6 - n1pn2) / gamma(17./6 - n1mn2
                                                      )/gamma(17./6 + n1mn2)
        temp2  = temp2 / gamma(n1pn2 + 23./6) 

        alpha  = [11./6, 7./3, 17./6]
        beta   = [11./6 - n1pn2, 17./6 - n1mn2, 17./6 + n1mn2, 23./6 + n1pn2]
        
        temp2  = temp2 * hyper(alpha, beta, x0**2)
        return fp.mpf(result * (temp1 + temp2))
    
    def get_covariance(self):
        if self._m1!=self._m2 or (self._j1 + self._j2) %2==1 and self._m1!=0:
            return 0.
        else:
            return self._compute_covariance()


import numpy as np
from arte.utils.rbf_generator import RBFGenerator
from arte.types.mask import CircularMask
from arte.types.wavefront import Wavefront
from arte.utils.base_modal_decomposer import BaseModalDecomposer


class RadialBasisModalDecomposer(BaseModalDecomposer):

    # write class and method documentation here
    """
    This class decomposes a wavefront into a set of modal coefficients. The modal coefficients can be
    Zernike, KL or RBF coefficients. The class provides methods to measure the modal coefficients from
    slopes or wavefronts. The methods are:
    - measureZernikeCoefficientsFromSlopes
    - measureZernikeCoefficientsFromWavefront
    - measureKLCoefficientsFromWavefront
    - measureModalCoefficientsFromWavefront
    Parameters
    ----------
    n_modes: int
        Number of modes to decompose the wavefront into.
    """

    def __init__(self, coordinates_list):
        self.coordinates_list = coordinates_list
        super().__init__(n_modes = len(coordinates_list))

    def generator(self, nModes, circular_mask, user_mask, rbfFunction=None, **kwargs):
        rbf = RBFGenerator(circular_mask, self.coordinates_list, rbfFunction=rbfFunction)
        rbf.generate()
        return rbf




#!/usr/bin/env python
'''

# who       when        what
# -------  ----------  ---------------------------------
# apuglisi  2022-02-30  Fixed bug in detecting downsampling
#                       on just one axis
# apuglisi  2020-04-15  Added sample option, added exceptions,
#                       allow any kind of sequence for new_shape
#           2012-11-20  Created from git://gist.github.com/1348792.git
'''
import numpy as np


def rebin(a, new_shape, sample=False):
    """
    Replacement of IDL's rebin() function for 2d arrays.

    Resizes a 2d array by averaging or repeating elements.
    New dimensions must be integral factors of original dimensions,
    otherwise a ValueError exception will be raised.

    Parameters
    ----------
    a : ndarray
        Input array.
    new_shape : 2-elements sequence
        Shape of the output array
    sample : bool
        if True, when reducing the array side elements are set
        using a nearest-neighbor algorithm instead of averaging.
        This parameter has no effect when enlarging the array.

    Returns
    -------
    rebinned_array : ndarray
        If the new shape is smaller of the input array  the data are averaged,
        unless the sample parameter is set.
        If the new shape is bigger array elements are repeated.

    Raises
    ------
    ValueError
        in the following cases:
         - new_shape is not a sequence of 2 values that can be converted to int
         - new dimensions are not an integral factor of original dimensions
    NotImplementedError
         - one dimension requires an upsampling while the other requires
           a downsampling

    Examples
    --------
    >>> a = np.array([[0, 1], [2, 3]])
    >>> b = rebin(a, (4, 6)) #upsize
    >>> b
    array([[0, 0, 0, 1, 1, 1],
           [0, 0, 0, 1, 1, 1],
           [2, 2, 2, 3, 3, 3],
           [2, 2, 2, 3, 3, 3]])
    >>> rebin(b, (2, 3)) #downsize
    array([[0. , 0.5, 1. ],
           [2. , 2.5, 3. ]])
    >>> rebin(b, (2, 3), sample=True) #downsize
    array([[0, 0, 1],
           [2, 2, 3]])
    """

    # unpack early to allow any 2-length type for new_shape
    m, n = map(int, new_shape)

    if a.shape == (m, n):
        return a

    M, N = a.shape

    if m <= M and n <= N:
        if (M//m != M/m) or (N//n != N/n):
            raise ValueError('Cannot downsample by non-integer factors')

    elif M <= m and M <= m:
        if (m//M != m/M) or (n//N != n/N):
            raise ValueError('Cannot upsample by non-integer factors')

    else:
        raise NotImplementedError('Up- and down-sampling in different axes '
                                  'is not supported')

    if sample:
        slices = [slice(0, old, float(old) / new)
                  for old, new in zip(a.shape, (m, n))]
        idx = np.mgrid[slices].astype(np.int32)
        return a[tuple(idx)]
    else:
        if m <= M and n <= N:
            return a.reshape((m, M//m, n, N//n)).mean(3).mean(1)
        elif m >= M and n >= M:
            return np.repeat(np.repeat(a, m/M, axis=0), n/N, axis=1)

# __oOo__




import time

class Timer(object):
    '''
    Timer context manager. Originally from
    https://stackoverflow.com/questions/5849800/tic-toc-functions-analog-in-python

    and modified to support multiple time units

    Example:

    with Timer(unit='s', msg='foo', precision=2):
        long_function()

    result:
    foo: Elapsed: 33.20 s
    '''
    def __init__(self, unit='s', precision=2, msg=''):
        self.unit = unit
        self.precision = precision
        self.msg = msg
        if unit not in ['s', 'ms', 'us']:
            raise Exception('Unit %s is not supported' % unit)

    def __enter__(self):
        self.t0 = time.time()

    def __exit__(self, type, value, traceback):
        elapsed = time.time() - self.t0

        name = self.name
        if name != '':
            name = ' (%s)' % name

        fmt = 'Elapsed time (%s): %%.%df %s' % (self.msg, self.precision, self.unit)

        number = {'s': elapsed,
                  'ms': elapsed * 1000,
                  'us': elapsed * 1000000 }

        print(fmt % number)





import numpy as np
from functools import cache
from scipy.special import factorial
from arte.types.mask import CircularMask


class ZernikeGenerator(object):
    '''
    Generator of Zernike polynomials and their derivatives

    The generator returns masked arrays representing Zernike polynomials
    sampled over the array grid. The pixels whose distance from the unit disk
    center is greater than the mask radius are masked.

    The discretization of the continous unit disk space into the array grid is
    done in the pixels' center (e.g. if the disk diameter is 2 pixels and the
    center is in x=1 then the first pixel contains polynomial values
    evaluated in x=-0.5 and the second pixel contains values evaluated in x=0.5)

    The user can specify the diameter of the pupil `pupil` or 
    specify a `~arte.types.mask.CircularMask`

    In the case a scalar real value is passed as `pupil` argument,
    The shape of the returned array cannot be specified: it is a 
    square array of size ceil(pupil_diameter).
    The center of the unit disk cannot be specified and it is the central pixel
    of the array (if the array size is odd) or the corner of the
    4 central pixels (if the array size is even).


    In the case a `~arte.types.mask.CircularMask` is passed as `pupil` argument,
    the mask is used as unit disk over which the polynomial are computed. Non-integer 
    center coordinates and radius are properly managed.



    Parameters
    ----------
        pupil: real or `~arte.types.mask.CircularMask`
            If a scalar value, the argument is used as pupil diameter in pixels.
            If a `~arte.types.mask.CircularMask`, the argument is used as mask
            representing the unit disk.


    Notes
    -----
    Polynomials normalization and ordering follows the conventions
    described in Noll's paper [1]_

    .. [1] Noll, R. J., “Zernike polynomials and atmospheric
       turbulence.”, Journal of the Optical Society of America
       (1917-1983), vol. 66, pp. 207–211, 1976.


    Examples
    --------

    Create a Zernike polynomial sampled with unit circle defined over 
    64 pixels representing tilt

    >>> zg = ZernikeGenerator(64)
    >>> tip = zg.getZernike(2)
    >>> tilt = zg.getZernike(3)
    >>> tilt = zg[3] # equivalent to getZernike(3) 


    '''

    def __init__(self, pupil):
        if isinstance(pupil, CircularMask):
            self._radius = pupil.radius()
            self._shape = pupil.shape()
            self._center = pupil.center()
            self._boolean_mask = pupil.mask()
        else:
            self._radius = pupil / 2
            sz = np.ceil(pupil)
            self._shape = (sz, sz)
            self._center = np.ones(2) * (sz / 2)
            cm = CircularMask(
                self._shape, maskCenter=self._center, maskRadius=self._radius)
            self._boolean_mask = cm.mask()

        self._rhoMap, self._thetaMap = self._polar_array()
        self._dx = None
        self._dy = None
        self._dictCache = {}
        self._dictDxCache = {}
        self._dictDyCache = {}

    def radius(self):
        '''
        Radius of the unit disk 

        Returns
        -------
        radius: real
            unit disk radius in pixel

        '''
        return self._radius

    def center(self):
        '''
        Y, X coordinates of the unit disk center

        Returns
        -------
        center: `~numpy.array` of shape (2,)
            Y, X coordinate of the unit disk center in the array reference system

        '''
        return self._center

    def first_mode(self):
        '''
        Index of first mode, excluding piston
        '''
        return 2

    def _derivativeCoeffX(self, index):
        if (self._dx is None) or (self._dx.shape[0] < index):
            self._dx = self._computeDerivativeCoeffX(index)
        return self._dx[0:index, 0:index]

    def _derivativeCoeffY(self, index):
        if (self._dy is None) or (self._dy.shape[0] < index):
            self._dy = self._computeDerivativeCoeffY(index)
        return self._dy[0:index, 0:index]

    @classmethod
    def degree(cls, index):
        n = cls.radial_order(index)
        cn = n * (n + 1) / 2 + 1
        if n % 2 == 0:
            m = int(index - cn + 1) // 2 * 2
        else:
            m = int(index - cn) // 2 * 2 + 1
        radialDegree = n
        azimuthalFrequency = m
        return radialDegree, azimuthalFrequency

    def _rnm(self, radialDegree, azimuthalFrequency, rhoArray):
        n = radialDegree
        m = azimuthalFrequency
        rho = rhoArray
        if (n - m) % 2 != 0:
            raise Exception("n-m must be even. Got %d-%d" % (n, m))
        if abs(m) > n:
            raise Exception("The following must be true |m|<=n. Got %d, %d" %
                            (n, m))

        if(n == 0 and m == 0):
            return np.ones(rho.shape)
        rho = np.where(rho < 0, 0, rho)
        Rnm = np.zeros(rho.shape)
        S = (n - abs(m)) // 2
        for s in range(0, S + 1):
            CR = pow(-1, s) * factorial(n - s) / \
                (factorial(s) * factorial(-s + (n + abs(m)) / 2) *
                 factorial(-s + (n - abs(m)) / 2))
            p = CR * pow(rho, n - 2 * s)
            Rnm = Rnm + p
        return Rnm

    def _polar(self, index, rhoArray, thetaArray):
        n, m = self.degree(index)
        rho = rhoArray
        theta = thetaArray

        Rnm = self._rnm(n, m, rho)
        NC = np.sqrt(2 * (n + 1))
        if m == 0:
            return np.sqrt(0.5) * NC * Rnm
        if index % 2 == 0:
            return NC * Rnm * np.cos(m * theta)
        else:
            return NC * Rnm * np.sin(m * theta)

    def cartesian_coordinates(self):
        '''
        Return X, Y maps of cartesian coordinates

        The Zernike polynomials and derivatives are evaluated over a grid of 
        points, whose coordinates are accessible through `cartesian_coordinates`

        Returns
        -------
        X,Y: `~numpy.array`
            coordinates of points of the unit disk where the polynomials
            are evaluated

        Examples
        --------

        Create Zernike polynomials on a unit disk defined over 
        4 pixels. The returned arrays have (4,4) shape (corresponding 
        to the coords [-0.75, -0.25, 0.25, 0.75])

        >>> zg = ZernikeGenerator(4)
        >>> x, y = zg.cartesian_coordinates()
        >>> x[0]
        array([-0.75, -0.25,  0.25,  0.75])

        As above with 3 pixels

        >>> zg = ZernikeGenerator(3)
        >>> x, y = zg.cartesian_coordinates()
        >>> x[0]
        array([-0.666, 0,  0.666])

        In case of non-integer diameter, the array size is rounded up to the
        next integer

        >>> zg = ZernikeGenerator(2.5)
        >>> x, y = zg.cartesian_coordinates()
        >>> x[0]
        array([-0.8, 0,  0.8])

        '''
        nPxY = self._shape[0]
        nPxX = self._shape[1]
        c = np.array(self.center())
        cc = np.expand_dims(c, axis=(1, 2))
        Y, X = (np.mgrid[0.5: nPxY + 0.5: 1,
                         0.5: nPxX + 0.5: 1] - cc) / self.radius()

        return X, Y

    def _polar_array(self):
        X, Y = self.cartesian_coordinates()
        r = np.sqrt(X ** 2 + Y ** 2)
        th = np.arctan2(Y, X)
        return r, th

    def getZernikeDict(self, indexVector):
        ret = {}
        for index in indexVector:
            ret[index] = self.getZernike(index)
        return ret

    def getModesDict(self, indexVector):
        return self.getZernikeDict(indexVector)

    def getZernike(self, index):
        if not self._is_integer_num(index):
            raise ValueError("Invalid Zernike index %s" % index)
        if index <= 0:
            raise ValueError("Invalid Zernike index %d" % index)
        if index not in list(self._dictCache.keys()):
            res = self._polar(index, self._rhoMap,
                              self._thetaMap)
            self._dictCache[index] = np.ma.masked_array(
                data=res, mask=self._boolean_mask)
        return self._dictCache[index]

    @staticmethod
    def _is_integer_num(n):
        if isinstance(n, (int, np.integer)):
            return True
        if isinstance(n, float):
            return n.is_integer()
        return False

    def __getitem__(self, index):
        return self.getZernike(index)

    def _computeDerivativeCoeffX(self, index):
        jmax = index
        G_mat = np.zeros((jmax, jmax))
        for i in range(1, jmax + 1):
            for j in range(1, jmax + 1):
                ni, mi = ZernikeGenerator.degree(i)
                nj, mj = ZernikeGenerator.degree(j)
                if (
                    (
                        (
                            (
                                mi != 0 and mj != 0
                            ) and (
                                (
                                    _isEven(i) and _isEven(j)
                                ) or (
                                    _isOdd(i) and _isOdd(j)
                                )
                            )
                        ) or (
                            (
                                (mi == 0) and _isEven(j)
                            ) or (
                                (mj == 0) and _isEven(i)
                            )
                        )
                    ) and (
                        (mj == mi + 1) or (mj == mi - 1)
                    ) and (
                        j < i
                    )
                ):
                    G_mat[i - 1, j - 1] = np.sqrt((ni + 1) * (nj + 1))
                    if ((mi == 0) or (mj == 0)):
                        G_mat[i - 1, j - 1] *= np.sqrt(2)
        return G_mat

    def _computeDerivativeCoeffY(self, index):
        jmax = index
        G_mat = np.zeros((jmax, jmax))
        for i in range(1, jmax + 1):
            for j in range(1, jmax + 1):
                ni, mi = ZernikeGenerator.degree(i)
                nj, mj = ZernikeGenerator.degree(j)
                if (
                    (
                        (
                            (
                                mi != 0 and mj != 0
                            ) and (
                                (
                                    _isOdd(i) and _isEven(j)
                                ) or (
                                    _isEven(i) and _isOdd(j)
                                )
                            )
                        ) or (
                            (
                                (mi == 0) and _isOdd(j)
                            ) or (
                                (mj == 0) and _isOdd(i)
                            )
                        )
                    ) and (
                        (mj == mi + 1) or (mj == mi - 1)
                    ) and (
                        j < i
                    )
                ):
                    G_mat[i - 1, j - 1] = np.sqrt((ni + 1) * (nj + 1))
                    if ((mi == 0) or (mj == 0)):
                        G_mat[i - 1, j - 1] *= np.sqrt(2)
                    if (
                        (
                            (
                                (mj == mi + 1) and _isOdd(i)
                            ) or (
                                (mj == mi - 1) and _isEven(i)
                            )
                        ) and (
                            mi != 0
                        )
                    ):
                        G_mat[i - 1, j - 1] *= -1

        return G_mat

    def getDerivativeXDict(self, indexVector):
        ret = {}
        for index in indexVector:
            ret[index] = self.getDerivativeX(index)
        return ret

    def getDerivativeYDict(self, indexVector):
        ret = {}
        for index in indexVector:
            ret[index] = self.getDerivativeY(index)
        return ret

    def getDerivativeX(self, index):
        if index not in self._dictDxCache:
            self._dictDxCache[index] = self._computeDerivativeX(index)
        return self._dictDxCache[index]

    def _computeDerivativeX(self, index):
        coeffX = self._derivativeCoeffX(index)
        dx = self.getZernike(1) * 0.
        for i in range(1, index):
            dx += coeffX[index - 1, i - 1] * self.getZernike(i)
        return dx

    def getDerivativeY(self, index):
        if index not in self._dictDyCache:
            self._dictDyCache[index] = self._computeDerivativeY(index)
        return self._dictDyCache[index]

    def _computeDerivativeY(self, index):
        coeffY = self._derivativeCoeffY(index)
        dy = self.getZernike(1) * 0.
        for i in range(1, index):
            dy += coeffY[index - 1, i - 1] * self.getZernike(i)
        return dy

    @classmethod
    def radial_order(cls, j):
        '''
        Return radial order of j-th polynomial

        Parameters
        ----------
        j: int or sequence of int
            polynomial index

        Returns
        -------
        n: int or sequence of int
            radial order of the specified indexes

        '''
        return np.ceil(0.5 * (np.sqrt(8 * np.array(j) + 1) - 3)).astype(int)

    @classmethod
    @cache
    def index_to_name_dict(cls):
        '''
        Zernike index to zernike name mapping

        Returns
        -------
        names: dict
            dictionary of Zernike mode names: int -> str
        '''
        return {1:'piston',
                2:'tip',
                3:'tilt',
                4:'focus'}

    @classmethod
    @cache
    def name_to_index_dict(cls):
        '''
        Zernike name to zernike index mapping

        Returns
        -------
        names: dict
            dictionary of Zernike indexes: str -> int
        '''
        return {v: k for k,v in cls.index_to_name_dict().items()}


def _isOdd(num):
    return num % 2 != 0


def _isEven(num):
    return num % 2 == 0



from functools import wraps
import time
import traceback
import inspect
import threading
import types
from numpy import ndarray

from arte.utils.logger import LoggerException


class ReturnTypeMismatchError(Exception):
    pass


def _raiseReturnTypeMismatchError(expctedType, actualType):
    raise ReturnTypeMismatchError(
        "Type mismatch: "
        "Expected is %s but got %s" % (expctedType, actualType))


def returns(returnType):

    def realDecorator(f):

        @wraps(f)
        def wrapperMethod(*args, **kwds):
            result = f(*args, **kwds)
            if not isinstance(result, returnType):
                resultType = type(result)
                _raiseReturnTypeMismatchError(returnType, resultType)
            return result

        return wrapperMethod

    return realDecorator


def returnsNone(f):
    return returns(type(None))(f)


def returnsForExample(exampleInstance):

    def realDecorator(f):

        @wraps(f)
        def wrapperMethod(*args, **kwds):
            result = f(*args, **kwds)
            resultType = type(result)
            exampleType = type(exampleInstance)
            if resultType != exampleType:
                _raiseReturnTypeMismatchError(exampleType, resultType)
            return result

        return wrapperMethod

    return realDecorator


def suppressException(resultInCaseOfFailure=None):

    def decorate(f):

        @wraps(f)
        def wrapper(self, *args, **kwds):
            try:
                return f(self, *args, **kwds)
            except Exception as e:
                self._logger.error(str(e))
                traceback.print_exc()
                return resultInCaseOfFailure

        return wrapper

    return decorate


def _logEnterAndExit(loggerMethod, enterMessage, exitMessage,
                     f, self, *args, **kwds):
    loggerMethod(enterMessage)
    res = f(self, *args, **kwds)
    loggerMethod(exitMessage)
    return res


def logEnterAndExit(enterMessage, exitMessage, level='notice'):

    def wrapperFunc(f):

        @wraps(f)
        def wrapper(self, *args, **kwds):
            if self._logger is None:
                raise LoggerException(
                    "Logger unavailable for message '%s' '%s'" %
                    (enterMessage, exitMessage))
            loggerMethod = self._logger.__getattribute__(level)
            return _logEnterAndExit(loggerMethod,
                                    enterMessage, exitMessage,
                                    f, self, *args, **kwds)

        return wrapper

    return wrapperFunc


def logTime(f):

    @wraps(f)
    def wrappedMethod(self, *args, **kwds):
        t0 = time.time()
        try:
            return f(self, *args, **kwds)
        finally:
            diffSec = time.time() - t0
            self._logger.notice("Method '%s' took %.3f sec" % (
                f.__name__, diffSec))

    return wrappedMethod


def cacheResult(f):

    @wraps(f)
    def wrapper(self, *args):
        cacheName = f.__name__ + "_cached_result"
        if cacheName not in self.__dict__:
            self.__dict__[cacheName] = {}

        key = ()
        for aa in args:
            if isinstance(aa, ndarray):
                key += (hash(aa.tostring()),)
            else:
                try:
                    key += (hash(aa),)
                except Exception:
                    key += (aa,)
        if key not in self.__dict__[cacheName]:
            result = f(self, *args)
            self.__dict__[cacheName][key] = result
        return self.__dict__[cacheName][key]

    return wrapper


def override(f):
    return f


def logFailureAndContinue(func):

    @wraps(func)
    def wrappedMethod(self, *args, **kwds):
        try:
            return func(self, *args, **kwds)
        except Exception as e:
            traceback.print_exc()
            self._logger.error("'%s' failed: %s" % (
                func.__name__, str(e)))

    return wrappedMethod


def _synchronizedWith(lock):

    def decorator(func):

        @wraps(func)
        def synchedFunc(*args, **kwds):
            with lock:
                return func(*args, **kwds)

        return synchedFunc

    return decorator


def _synchronizedWithAttr(lockName):

    def decorator(method):

        @wraps(method)
        def synchronizedMethod(self, *args, **kwds):
            lock = self.__dict__[lockName]
            with lock:
                return method(self, *args, **kwds)

        return synchronizedMethod

    return decorator


def synchronized(item):

    if isinstance(item, str):
        return _synchronizedWithAttr(item)
    elif inspect.isclass(item):
        syncClass = item
        lock = threading.RLock()

        origInit = syncClass.__init__

        def __init__(self, *args, **kwds):
            self.__lock__ = lock
            origInit(self, *args, **kwds)

        syncClass.__init__ = __init__

        for key in syncClass.__dict__:
            val = syncClass.__dict__[key]
            if isinstance(val, types.FunctionType):
                decorator = _synchronizedWith(lock)
                setattr(syncClass, key, decorator(val))

        return syncClass
    else:
        assert False, "Unsupported item type: %s is of type %s" % (
            str(item), type(item))



from warnings import warn
from arte.types.zernike_coefficients import ZernikeCoefficients
from arte.utils.decorator import returns
from arte.utils.zernike_decomposer import ZernikeModalDecomposer


class ModalDecomposer(ZernikeModalDecomposer):
    '''
    Backward compatibility with old ModalDecomposer class
    '''
    def __init__(self, n_modes=None, n_zernike_modes=None):
        if n_zernike_modes is not None:
            n_modes = n_zernike_modes
        if n_modes is None and n_zernike_modes is None:
            raise ValueError("either n_modes or n_zernike_modes must be specified")
        super().__init__(n_modes)
        warn(
            f"{self.__class__.__name__} will be deprecated or change use. Use ZernikeModalDecomposer instead",
            DeprecationWarning,
            stacklevel=2,
        )

    @returns(ZernikeCoefficients)
    def measureZernikeCoefficientsFromWavefront(self, wavefront, circular_mask,
                                                user_mask=None, nModes=None, dtype=float):
        return self.measureModalCoefficientsFromWavefront(wavefront, circular_mask,
                                                       user_mask, nModes, dtype=dtype)

    @returns(ZernikeCoefficients)
    def measureZernikeCoefficientsFromSlopes(self, slopes, circular_mask,
                                             user_mask=None, nModes=None, dtype=float):
        return self.measureModalCoefficientsFromSlopes(slopes, circular_mask,
                                                       user_mask, nModes, dtype=dtype)

    def synthZernikeRecFromSlopes(self, nModes, circular_mask, user_mask=None, dtype=float):
        return self.cachedSyntheticReconstructorFromSlopes(nModes, circular_mask,
                                                           user_mask, dtype=dtype)

    def synthZernikeRecFromWavefront(self, nModes, circular_mask, user_mask=None, dtype=float):
        return self.cachedSyntheticReconstructorFromWavefront(nModes, circular_mask,
                                                           user_mask, dtype=dtype)



# -*- coding: utf-8 -*-

from itertools import tee


def flatten(x):
    '''
    Generator that flatten arbitrarily nested lists.

    This generator will flatten a list that may contain
    other lists (nested arbitrarily) and simple items
    into a flat list.

    >>> flat = flatten([[1,[2,3]],4,[5,6]])
    >>> list(flat)
    [1, 2, 3, 4, 5, 6]
    '''
    for item in x:
        try:
            yield from flatten(item)
        except TypeError:
            yield item


def pairwise(iterable):
    '''
    Pairwise iterator, from itertool recipes.

    See Python library docs, itertools chapter.

    >>> s = [0,1,2,3,4,5,6]
    >>> list(pairwise(s))
    [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 6)]
    '''
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)



import datetime
import functools

from arte.utils.help import add_help
from arte.dataelab.cache_on_disk import set_tag, clear_cache


class PostInitCaller(type):
    '''Meta class with a _post_init() method.
    
    Used to initialize the disk cache, since DiskCacher objects are
    available only after all child/member objects have been initialized'''
    def __call__(cls, *args, **kwargs):
        obj = type.__call__(cls, *args, **kwargs)
        obj._post_init()
        return obj


@add_help
class BaseAnalyzer(metaclass=PostInitCaller):
    '''Main analyzer object
    
    Note that an analyzer is not CanBeIncomplete, because all attributes
    should be initialized (if they are missing, their value will be NA)
    '''

    def __init__(self, snapshot_tag, recalc=False):
        self._snapshot_tag = snapshot_tag
        self._recalc = recalc
 
    def _post_init(self):
        '''Initialize disk cache
        
        Executed after all child classes have completed their __init__
        thanks to the PostInitCaller metaclass
        '''
        set_tag(self, self._snapshot_tag)
        if self._recalc:
            self.recalc()
            self._recalc = False

    @classmethod
    def get(cls, tag, *args, recalc=False, **kwargs):
        '''Get the Analyzer instance (or derived class) corresponding to *tag*.

        This method mantains an internal cache. If a tag is requested
        multiple times, the same Analyzer instance is returned.
        
        Parameters
        ----------
        tag: str
            snapshot tag
        recalc: bool, optional
            if set to True, any cached data for this tag will be deleted and
            computed again when requested
         '''
        analyzer = cls._get(tag, *args, **kwargs)
        # Special recalc handling, must be set again for a cached instance
        if recalc:
            analyzer.recalc()
        return analyzer

    @classmethod
    @functools.cache
    def _get(cls, tag, *args, recalc=False, **kwargs):
        '''Get a new Analyzer instance 

        Added one level of indirection (get() calls _get())
        to make sure that the cache always works even when
        the default argument *recalc* is not specified in get().

        Also makes it easier to override it in classes
        '''
        return cls(tag, *args, recalc=recalc, **kwargs)

    def recalc(self):
        '''Force recalculation of this analyzer data'''
        clear_cache(self)

    def snapshot_tag(self):
        '''Snapshot tag for this Analyzer object'''
        return self._snapshot_tag

    def date_in_seconds(self):
        '''Tag date as seconds since the epoch'''
        epoch = datetime.datetime(1970, 1, 1, 0, 0, 0, 0)
        this_date = datetime.datetime(int(self._snapshot_tag[0:4]),
                                     int(self._snapshot_tag[4:6]),
                                     int(self._snapshot_tag[6:8]),
                                     int(self._snapshot_tag[9:11]),
                                     int(self._snapshot_tag[11:13]),
                                     int(self._snapshot_tag[13:15]))
        td = this_date - epoch
        return (td.microseconds + (td.seconds + td.days * 24 * 3600) * 1e6) / 1e6

    # Override to add additional info
    def _info(self):
        return {'snapshot_tag': self._snapshot_tag}

    def info(self):
        '''Info dictionary'''
        return self._info()

    def summary(self, keywidth=None):
        '''Print info dictionary on stdout'''
        info = self._info()
        if keywidth is None:
            keywidth = max(len(x) for x in info.keys())
        for k, v in info.items():
            print(f'{k:{keywidth}} : {v}')

    def wiki(self, header=True):
        '''Print info dictionary in wiki format on stdout'''
        info = self._info()
        spacing = [max([len(x)+2, len(info[x])]) for x in info.keys()]

        if header:
            for i, k in enumerate(info.keys()):
                print(f'|*{k:^{spacing[i]}}*', end='')
            print('|')

        for i, v in enumerate(info.values()):
            print(f'|{v:^{spacing[i]}}', end='')
        print('|')




import numpy as np
from arte.utils.discrete_fourier_transform \
    import BidimensionalFourierTransform as bfft

from astropy.io import fits

class PhaseScreenGenerator(object):

    def __init__(self,
                 screenSizeInPixels,
                 screenSizeInMeters,
                 outerScaleInMeters,
                 seed=None):
        self._screenSzInPx = screenSizeInPixels
        self._screenSzInM = float(screenSizeInMeters)
        self._outerScaleInM = float(outerScaleInMeters)
        self._phaseScreens = None
        self._nSubHarmonicsToUse = 8
        if seed is None:
            self._seed = np.random.randint(2**32 - 1, dtype=np.uint32)
        else:
            self._seed = seed

    def _spatial_frequency(self, screenSizeInPixels):
        a = np.tile(np.fft.fftfreq(screenSizeInPixels, d=1. / screenSizeInPixels),
                    (screenSizeInPixels, 1))
        return np.linalg.norm(np.dstack((a, a.T)), axis=2)

    def generate_normalized_phase_screens(self, numberOfScreens):
        np.random.seed(self._seed)
        nIters = int(np.ceil(numberOfScreens / 2))
        ret = np.zeros((2 * nIters, self._screenSzInPx, self._screenSzInPx))
        for i in range(nIters):
            ps = self._generate_phase_screen_with_fft()
            ps += self._generate_sub_harmonics(self._nSubHarmonicsToUse)
            ret[2* i, :, :]= self._remove_piston(np.sqrt(2)* ps.real)
            ret[2* i + 1, :, :]= self._remove_piston(np.sqrt(2)* ps.imag)
        self._phaseScreens= ret[:numberOfScreens]


    def _remove_piston(self, scrn):
        return scrn-scrn.mean()


    def _generate_phase_screen_with_fft(self):
        '''
        Normalized to 1pix/r0
        '''
        freqMap = self._spatial_frequency(self._screenSzInPx)
        modul = self._kolmogorov_amplitude_map_no_piston(freqMap)
        phaseScreen = np.sqrt(0.0228) * self._screenSzInPx**(5. / 6) * \
            np.fft.fft2(modul * np.exp(self._random_phase() * 1j))
        return phaseScreen

    def rescale_to(self, r0At500nm):
        self._normalizationFactor = \
            ((self._screenSzInM / self._screenSzInPx) / r0At500nm) ** (5. / 6)

    def get_in_radians_at(self, wavelengthInMeters):
        return 500e-9 / wavelengthInMeters * \
            self._normalizationFactor * self._phaseScreens

    def get_in_meters(self):
        return self._normalizationFactor * self._phaseScreens / \
            (2 * np.pi) * 500e-9

    def _random_phase(self):
        return np.random.rand(self._screenSzInPx, self._screenSzInPx) * 2 * np.pi

    def _kolmogorov_amplitude_map_no_piston(self, freqMap):
        mappa = (freqMap**2 + (self._screenSzInM / self._outerScaleInM)**2)**(
            -11. / 12)
        mappa[0, 0] = 0
        return mappa

    def _generate_sub_harmonics(self, numberOfSubHarmonics):
        nSub = 3
        lowFreqScreen = np.zeros((self._screenSzInPx, self._screenSzInPx),
                                 dtype=np.complex128)
        freqX = bfft.frequencies_x_map(nSub, 1. / nSub)
        freqY = bfft.frequencies_y_map(nSub, 1. / nSub)
        freqMod = bfft.frequencies_norm_map(nSub, 1. / nSub)
        vv = np.arange(self._screenSzInPx) / self._screenSzInPx
        xx = np.tile(vv, (self._screenSzInPx, 1))
        yy = xx.T
        depth = 0
        while depth < numberOfSubHarmonics:
            depth += 1
            phase = self._random_phase()
            freqMod /= nSub
            freqX /= nSub
            freqY /= nSub
            modul = self._kolmogorov_amplitude_map_no_piston(freqMod)
            for ix in range(nSub):
                for jx in range(nSub):
                    sh = np.exp(2 * np.pi * 1j *
                                (xx * freqX[ix, jx] + yy * freqY[ix, jx] + phase[ix, jx]))
                    sh0 = sh.sum() / self._screenSzInPx**2
                    lowFreqScreen += 1. / nSub**depth * \
                        modul[ix, jx] * (sh - sh0)
        lowFreqScreen *= np.sqrt(0.0228) * self._screenSzInPx**(5. / 6)
        return lowFreqScreen

    def _determineNumberOfSubHarmonics(self):
        maxNoOfSubHarm = 8
        pass
    
    def save_normalized_phase_screens(self, fname):
        hdr = fits.Header()
        hdr['SZ_IN_PX'] =  self._screenSzInPx
        hdr['SZ_IN_M'] = self._screenSzInM
        hdr['OS_IN_M'] = self._outerScaleInM
        hdr['SEED'] = self._seed
        
        fits.writeto(fname, self._phaseScreens, hdr)
        
    
    @staticmethod 
    def load_normalized_phase_screens(fname):
        header = fits.getheader(fname)
        screenSizeInPixels = header['SZ_IN_PX'] 
        screenSizeInMeters = header['SZ_IN_M'] 
        outerScaleInMeters = header['OS_IN_M'] 
        seed = header['SEED'] 
        
        psg = PhaseScreenGenerator(
            screenSizeInPixels,
            screenSizeInMeters,
            outerScaleInMeters,
            seed)
        
        hduList = fits.open(fname)
        psg._phaseScreens = hduList[0].data
        
        return psg





def _data_root_dir_legacy():
    import pkg_resources

    dataroot = pkg_resources.resource_filename(
        'arte',
        'data')
    return dataroot

def dataRootDir():
    try:
        import importlib.resources
        dataroot = importlib.resources.files('arte') / 'data'
    except AttributeError:
        dataroot = _data_root_dir_legacy()
    return dataroot

